# coding=utf-8
from collections import OrderedDict

import numpy as np
from artemis.general.mymath import softmax, cosine_distance
from artemis.general.should_be_builtins import remove_duplicates
from artemis.general.tables import build_table
from artemis.ml.datasets.datasets import DataSet
from artemis.ml.tools.iteration import zip_minibatch_iterate_info, IterationInfo, minibatch_process

__author__ = 'peter'

"""
All evaluation functions in here are of the form

score = evaluation_fcn(actual, target)

Where:
    score is a scalar
    actual is an (n_samples, ...) array
    target is an (n_samples, ....) array
"""


def train_online_predictor(predictor, training_set, minibatch_size, n_epochs = 1):
    """
    Train a predictor on the training set
    :param predictor: An IPredictor object
    :param training_set: A DataCollection object
    :param minibatch_size: An integer, or 'full' for full batch training
    :param n_epochs: Number of passes to make over the training set.
    """
    print 'Training Predictor %s...' % (predictor, )
    for (_, data, target) in training_set.minibatch_iterator(minibatch_size = minibatch_size, epochs = n_epochs, single_channel = True):
        predictor.train(data, target)
    print 'Done.'


def evaluate_predictor(predictor, test_set, evaluation_function):
    if isinstance(evaluation_function, str):
        evaluation_function = get_evaluation_function(evaluation_function)
    output = predictor.predict(test_set.input)
    score = evaluation_function(actual = output, target = test_set.target)
    return score



def get_evaluation_function(name):
    return {
        'mse': mean_squared_error,
        'mean_cosine_distance': lambda a, b: cosine_distance(a, b, axis=1).mean(),
        'mean_squared_error': mean_squared_error,
        'mean_l1_error': mean_l1_error,
        'percent_argmax_correct': percent_argmax_correct,
        'percent_argmax_incorrect': percent_argmax_incorrect,
        'percent_correct': percent_correct,
        'softmax_categorical_xe': softmax_categorical_xe
        }[name]


def mean_l1_error(actual, target):
    return np.mean(np.sum(np.abs(actual-target), axis=-1), axis=-1)


def mean_squared_error(actual, target):
    return np.mean(np.sum((actual-target)**2, axis = -1), axis = -1)


def softmax_categorical_xe(actual, target):
    """
    :param actual: An (n_samples, n_dims) array identifying pre-logistic output
    :param target: An (n_samples, ) integer array identifying labels
    :return:
    """
    if target.ndim==1:
        assert target.dtype==int and np.max(target) < actual.shape[1]
    elif target.ndim==2:
        assert np.all(target.sum(axis=1)==1) and np.all(np.max(target, axis=1)==1)
        target = np.argmax(target, axis=1)
    else:
        raise Exception("Don't know how to interpret a {}-D target".format(target))

    return np.mean(softmax(actual, axis=1)[np.arange(actual.shape[0]), target], axis=0)


def fraction_correct(actual, target):
    return np.mean(actual == target)


def percent_correct(actual, target):
    return 100*fraction_correct(actual, target)


def percent_argmax_correct(actual, target):
    """
    :param actual: An (n_samples, n_dims) array
    :param target: An (n_samples, ) array of indices OR an (n_samples, n_dims) array
    :return:
    """
    actual = collapse_onehot_if_necessary(actual)
    target = collapse_onehot_if_necessary(target)
    return 100*fraction_correct(actual, target)


def percent_binary_incorrect(actual, target):
    return 100.-percent_binary_correct(actual, target)


def percent_binary_correct(actual, target):
    """
    :param actual:  A (n_samples, ) array of floats between 0 and 1
    :param target: A (n_samples, ) array of True/False
    :return: The percent of times the "actual" was closes to the correct.
    """
    assert len(actual) == len(target)
    if actual.ndim==2:
        assert actual.shape[1]==1
        actual = actual[:, 0]
    if target.ndim==2:
        assert target.shape[1]==1
        target = target[:, 0]
    assert target.ndim==1
    if actual.ndim>1:
        assert actual.shape == (len(target), 1)
        actual = actual[:, 0]
    if np.array_equal(np.unique(target), (0, 1)):
        assert np.all(actual)>=0 and np.all(actual)<=1
        assert np.all((target==0)|(target==1))
        return 100*np.mean(np.round(actual) == target)
    elif np.array_equal(np.unique(target), (-1, 1)):
        assert np.all((target==-1)|(target==1))
        return 100*np.mean((actual>0)*2-1 == target)
    else:
        raise Exception("Go away I'm tired.")


def percent_argmax_incorrect(actual, target):
    return 100 - percent_argmax_correct(actual, target)


def collapse_onehot_if_necessary(output_data):
    """
    Given an input that could either be in onehot encoding or not, return it in onehot encoding.

    :param output_data: Either an (n_samples, n_dims) array, or an (n_samples, ) array of labels.
    :return: An (n_samples, ) array.
    """

    output_data = np.squeeze(output_data)

    if output_data.ndim == 2:
        return np.argmax(output_data, axis = 1)
    else:
        assert output_data.ndim == 1 and output_data.dtype in (int, 'int32', bool)
        return output_data


class ModelTestScore(object):
    """
    An object representing the evaluation of a model on
    - one or more test sets
    - one or more prediction functions
    - one or more costs
    """

    def __init__(self, ):
        self.scores = OrderedDict()

    def __getitem__(self, (data_subset, prediction_function_name, cost_name)):
        return self.scores[data_subset, prediction_function_name, cost_name]

    def __setitem__(self, (data_subset, prediction_function_name, cost_name), value):
        self.scores[data_subset, prediction_function_name, cost_name] = value

    def keys(self):
        return self.scores.keys()

    def values(self):
        return self.scores.values()

    def iteritems(self):
        return self.scores.iteritems()

    def get_score(self, subset=None, prediction_function=None, cost_name=None):
        if subset is None:
            subset = self.get_only_data_subset()
        if prediction_function is None:
            prediction_function = self.get_only_prediction_function()
        if cost_name is None:
            cost_name = self.get_only_cost()
        return self[subset, prediction_function, cost_name]

    def get_data_subsets(self):
        return remove_duplicates([s for s, _, _ in self.scores.keys()])

    def get_prediction_functions(self):
        return remove_duplicates([f for _, f, _ in self.scores.keys()])

    def get_costs(self):
        return remove_duplicates([c for _, c, c in self.scores.keys()])

    def get_only_data_subset(self):
        return self._get_only_element(self.get_data_subsets(), 'Data Subset')

    def get_only_prediction_function(self):
        return self._get_only_element(self.get_prediction_functions(), 'Prediction Function')

    def get_only_cost(self):
        return self._get_only_element(self.get_costs(), 'Cost')

    @staticmethod
    def _get_only_element(elements, category_name):
        if len(elements)==1:
            result, = elements
            return result
        else:
            raise Exception("You need to specify the {}.  Options are: {}".format(category_name, elements))

    def __str__(self):
        sections = [u'({}{}{})={}'.format(
            subset,
            '' if pred_fcn is None else ','+str(pred_fcn),
            ','+str(cost_fcn),
            self.get_score(subset, pred_fcn, cost_fcn))
            for subset, pred_fcn, cost_fcn in self.keys()
            ]
        return '{}:{}'.format(self.__class__.__name__, ','.join(sections))

    def get_table(self):
        # test_pair_names, function_names, cost_names = [remove_duplicates(k) for k in zip(*self.scores.keys())]

        def lookup( (test_pair_name_, function_name_), cost_name_):
            return self[test_pair_name_, function_name_, cost_name_]

        rows = build_table(
            lookup_fcn=lookup,
            row_categories=[[test_pair_name for test_pair_name in self.get_data_subsets()], [function_name for function_name in self.get_prediction_functions()]],
            column_categories=[cost_name for cost_name in self.get_costs()],
            row_header_labels=['Subset', 'Function'],
            clear_repeated_headers=False,
            remove_unchanging_cols=False
        )
        import tabulate
        return tabulate.tabulate(rows)


class InfoScorePair(object):

    def __init__(self, info, score):
        """
        :param info: An IterationInfo Object
        :param v: A ModelTestScore object
        """
        assert isinstance(info, IterationInfo)
        assert isinstance(score, ModelTestScore)
        self.info = info
        self.score = score

    def __iter__(self):
        return iter([self.info, self.score])

    def __str__(self):
        return 'Epoch {} (after {:.3g}s)\n{}'.format(self.info.epoch, self.info.time, self.score)

    def get_table_headers(self):
        iterfields = IterationInfo._fields
        subset_names, prediction_funcs, cost_funcs = zip(*self.score.keys())
        return [
            (' ', )*len(iterfields) + subset_names,
            (' ', )*len(iterfields) + prediction_funcs,
            iterfields + cost_funcs
            ]

    def get_table_row(self):
        return [v for v in self.info] + self.score.values()

    def get_table(self, remove_headers = False):
        from tabulate import tabulate
        iterfields = IterationInfo._fields
        subset_names, prediction_funcs, cost_funcs = zip(*self.score.keys())
        headers = [
            (' ', )*len(iterfields) + subset_names,
            (' ', )*len(iterfields) + prediction_funcs,
            iterfields + cost_funcs
            ]
        data = [v for v in self.info] + self.score.values()
        table = tabulate(headers+[data], tablefmt='plain')
        if remove_headers:
            table = table[table.rfind('\n'):]

        return table


class InfoScorePairSequence(object):
    """
    An object representing a sequence of pairs of IterationInfo, Score objects.
    """

    def __init__(self):
        self._pairs = []

    def __iter__(self):
        return iter(self._pairs)

    def __getitem__(self, ix):
        return self._pairs[ix]

    def __len__(self):
        return len(self._pairs)

    def append(self, info_score_pair):
        self._pairs.append(info_score_pair)

    def get_oneliner(self, subset = 'test', prediction_function = None, score_measure = None, lower_is_better = False):
        """
        Return a 1-liner descibing the best score.
        """
        subset, prediction_function, score_measure = self._fill_fields(subset, prediction_function, score_measure)
        best_pair = self.get_best(subset, prediction_function, score_measure, lower_is_better=lower_is_better)
        return 'Best: Epoch {} of {}, {}: {}'.format(best_pair.info.epoch, self._pairs[-1].info.epoch, score_measure, best_pair.score[subset, prediction_function, score_measure])

    def get_values(self, subset = 'test', prediction_function = None, score_measure = None):
        return [score.get_score(subset, prediction_function, score_measure) for _, score in self]

    def get_best_value(self, subset = 'test', prediction_function = None, score_measure = None, lower_is_better = False):

        subset, prediction_function, score_measure = self._fill_fields(subset, prediction_function, score_measure)
        best_pair = self.get_best(subset, prediction_function, score_measure, lower_is_better=lower_is_better)
        try:
            return best_pair.score[subset, prediction_function, score_measure]
        except KeyError as err:
            raise KeyError(str(err.message) + '.  Indices must be ({{{}}}, {{{}}}, {{{}}}'.format(self.get_data_subsets(), self.get_prediction_functions(), self.get_costs()))

    def get_best(self, subset = 'test', prediction_function = None, score_measure = None, lower_is_better = False):
        """
        Given a list of (info, score) pairs which represet the progress over training, find the best score and return it.
        :param score_info_pairs: A list<(IterationInfo, dict)> of the type returned in train_and_test_online_predictor
        :param subset: 'train' or 'test' ... which subset to use to look for the best score
        :param prediction_function: Which prediction function (if there are multiple prediction functions, otherwise leave blank)
        :param score_measure: Which score measure (if there are multiple score measures, otherwise leave blank)
        :param lower_is_better: True if a lower score is better for the chosen score_measure
        :return: A InfoScorePair object
        """
        assert len(self._pairs) > 0, "You need to have at least one score to determine the best one."
        subset, prediction_function, score_measure = self._fill_fields(subset, prediction_function, score_measure)
        best_pair = self._pairs[0]
        for info_score_pair in self._pairs[1:]:
            best_pair = info_score_pair if (info_score_pair.score[subset, prediction_function, score_measure] < best_pair.score[subset, prediction_function, score_measure]) == lower_is_better else best_pair
        return best_pair

    def _fill_fields(self, subset = None, prediction_function = None, score_measure = None):
        first_score = self._pairs[0].score
        if subset is None:
            subset = first_score.get_only_data_subset()
        else:
            assert subset in self.get_data_subsets(), 'Subset {} is not one of {}'.format(subset, self.get_data_subsets())
        if prediction_function is None:
            prediction_function = first_score.get_only_prediction_function()
        else:
            assert prediction_function in self.get_prediction_functions(), 'Prediction function {} is not one of {}'.format(prediction_function, self.get_prediction_functions())
        if score_measure is None:
            score_measure = first_score.get_only_cost()
        else:
            assert score_measure in self.get_costs(), 'Score function {} is not one of {}'.format(score_measure, self.get_costs())
        return subset, prediction_function, score_measure

    def __str__(self):
        desc = 'InfoScorePairSequence<{} epochs over {}s. '.format(self[-1].info.epoch, self[-1].info.time)
        keys = self[0].score.keys()
        for subset, pred_fcn, cost_fcn in keys:
            values = self.get_values(subset, pred_fcn, cost_fcn)
            desc+=',({}{}{}) in [{} to {}]'.format(subset, '' if pred_fcn is None else ','+str(pred_fcn), ','+str(cost_fcn), min(values), max(values))
        desc += '>'
        return desc

    def get_table(self):
        from tabulate import tabulate
        if len(self)==0:
            return repr(self) + '\n' + '<Empty>'
        else:
            rows = self[0].get_table_headers() + [pair.get_table_row() for pair in self]
            return repr(self)+'\n  '+tabulate(rows).replace('\n', '\n  ')

    def get_summary(self, subset = None, prediction_function = None, score_measure = None, lower_is_better = False, score_format='.3g'):
        subsets = self._pairs[0].score.get_data_subsets() if subset is None else [subset]
        funcs = self._pairs[0].score.get_prediction_functions() if prediction_function is None else [prediction_function]
        costs = self._pairs[0].score.get_costs() if score_measure is None else [score_measure]
        entries = []
        for subset in subsets:
            for func in funcs:
                for cost in costs:
                    title = ','.join(([subset] if len(subsets)>1 else [])+([func] if len(funcs)>1 else [])+([cost] if len(costs)>1 else []))
                    score = self.get_best_value(subset=subset, prediction_function=func, score_measure=cost, lower_is_better=lower_is_better)
                    entries.append(('{}:{:'+score_format+'}').format(title, score))
        return '; '.join(entries)

    def get_data_subsets(self):
        return self._pairs[0].score.get_data_subsets()

    def get_prediction_functions(self):
        return self._pairs[0].score.get_prediction_functions()

    def get_costs(self):
        return self._pairs[0].score.get_costs()


def plot_info_score_pairs(ispc, prediction_function = None, score_measure=None, name='', show=True):
    """
    :param info_score_pair_sequences: An InfoScorePairSequence or a list of InfoScorePair Sequences, or a dict of them.
    :return:
    """

    from matplotlib import pyplot as plt

    colour = next(plt.gca()._get_lines.prop_cycler)['color']
    epochs = [info.epoch for info, _ in ispc]
    training_curve = ispc.get_values(subset='train', prediction_function = prediction_function, score_measure=score_measure)
    test_curve = ispc.get_values(subset='test', prediction_function=prediction_function, score_measure=score_measure)
    plt.plot(epochs, training_curve, color=colour, linestyle='--', label='{}:{}'.format(name, 'training'))
    plt.plot(epochs, test_curve, color=colour, label='{}:{}'.format(name, 'test'))
    plt.xlabel('Epoch')
    plt.ylabel('Score')
    plt.legend()
    plt.grid()


def plot_info_score_pairs_collection(info_score_pair_sequences, prediction_function = None, score_measure=None, show=True):

    from matplotlib import pyplot as plt
    if isinstance(info_score_pair_sequences, InfoScorePairSequence):
        info_score_pair_sequences = [info_score_pair_sequences]

    if isinstance(info_score_pair_sequences, (list, tuple)):
        info_score_pair_sequences = OrderedDict((ix, ispc) for ix, ispc in enumerate(info_score_pair_sequences))

    colours = [p['color'] for p in plt.rcParams['axes.prop_cycle']]
    for (name, ispc), colour in zip(info_score_pair_sequences.iteritems(), colours):
        plot_info_score_pairs(ispc, prediction_function = prediction_function, score_measure=score_measure, show=False, name=name)
    if show:
        plt.show()


def _dataset_to_test_pair(dataset, include = 'train+test'):
    """
    :param dataset:
    :param include:
    :return:
    """
    assert include in ('train', 'test', 'train+test', 'training', 'training+test')
    pairs = []
    if 'train' in include:
        pairs.append(('train', (dataset.training_set.input, dataset.training_set.target)))
    if 'test' in include:
        pairs.append(('test', (dataset.test_set.input, dataset.test_set.target)))
    return pairs


def assess_prediction_functions(test_pairs, functions, costs, print_results=False, prediction_minibatches = None):
    """

    :param test_pairs: A list<pair_name, (x, y)>, where x, y are equal-length vectors representing the samples in a dataset.
        Eg. [('training', (x_train, y_train)), ('test', (x_test, y_test))]
    :param functions: A list<function_name, function> of functions for computing the forward pass.
    :param costs: A list<(cost_name, cost_function)> or dict<cost_name: cost_function> of cost functions, where cost_function has the form:
        cost = cost_fcn(guess, y), where cost is a scalar, and guess is the output of the prediction function given one
            of the inputs (x) in test_pairs.
    :param prediction_minibatches: Size of minibatches to predict in.
    :return: A ModelTestScore object
    """
    if isinstance(test_pairs, DataSet):
        test_pairs = _dataset_to_test_pair(test_pairs)
    assert isinstance(test_pairs, list)
    assert all(len(_)==2 for _ in test_pairs)
    assert all(len(pair)==2 for name, pair in test_pairs)
    if isinstance(functions, dict):
        functions = functions.items()
    if callable(functions):
        functions = [(functions.__name__ if hasattr(functions, '__name__') else None, functions)]
    else:
        assert all(callable(f) for name, f in functions)
    if callable(costs):
        costs = [(costs.__name__, costs)]
    elif isinstance(costs, basestring):
        costs = [(costs, get_evaluation_function(costs))]
    elif isinstance(costs, dict):
        costs = costs.items()
    else:
        costs = [(cost, get_evaluation_function(cost)) if isinstance(cost, basestring) else (cost.__name__, cost) if callable(cost) else cost for cost in costs]
    assert all(callable(cost) for name, cost in costs)

    results = ModelTestScore()
    for test_pair_name, (x, y) in test_pairs:
        for function_name, function in functions:
            if prediction_minibatches is None:
                predictions = function(x)
            else:
                predictions = minibatch_process(function, minibatch_size=prediction_minibatches, mb_args=(x, ))
            for cost_name, cost_function in costs:
                results[test_pair_name, function_name, cost_name] = cost_function(predictions, y)

    if print_results:
        print results.get_table()

    return results


def print_score_results(score, info=None):
    """
    :param results: An OrderedDict in the format returned by assess_prediction_functions_on_generator.
    :return:
    """
    if info is not None:
        print 'Epoch {} (after {:.3g}s)'.format(info.epoch, info.time)
    test_pair_names, function_names, cost_names = [remove_duplicates(k) for k in zip(*score.keys())]
    rows = build_table(
        lookup_fcn=lambda (test_pair_name_, function_name_), cost_name_: score[test_pair_name_, function_name_, cost_name_],
        row_categories = [[test_pair_name for test_pair_name in test_pair_names], [function_name for function_name in function_names]],
        column_categories = [cost_name for cost_name in cost_names],
        row_header_labels=['Subset', 'Function'],
        clear_repeated_headers = False,
        remove_unchanging_cols=True
        )
    import tabulate
    print tabulate.tabulate(rows)


def train_and_test_online_predictor(dataset, train_fcn, predict_fcn, minibatch_size, n_epochs=None, test_epochs=None,
            score_measure='percent_argmax_correct', test_callback=None, training_callback = None, score_collection = None,
            test_on = 'training+test', pass_iteration_info_to_training = False, predict_in_minibatches=False):
    """
    Train an online predictor.  Return a data structure with info about the training.
    :param dataset: A DataSet object
    :param train_fcn: A function of the form train_fcn(x, y) which updates the parameters, OR
        train_fcn(x, y, iteration_info)   if pass_iteration_info_to_training==True
    :param predict_fcn: A function of the form y=predict_fcn(x) which makes a prediction giben inputs
    :param minibatch_size: Minibatch size
    :param n_epochs: Number of epoch
    :param test_epochs: Epochs to test at
    :param score_measure: String or function of the form:
        score = score_measure(guess, ground_truth)
        To be used in testing.
    :param test_callback: Function to be called after a test.  It has the form: f(info, score)
    :param training_callback: Function to be called after every training iteration.  It has the form f(info, x, y) where
    :param score_collection: If not None, a InfoScoreCollection object into which you save scores.  This allows you to
        access the scores before this function returns.
    :param test_on: What to run tests on: {'training', 'test', 'training+test'}
    :param pass_iteration_info_to_training: Pass an IterationInfo object into the training function (x, y, iter_info).
        This object contains fields (epoch, sample, time) which can be used to do things like adjust parameters on a schedule.
    :param predict_in_minibatches: Also run the prediction function in minibatches (for example to save memory)
    :return: A list<info, scores>  where...
        IterationInfo object (see artemis.ml.tools.iteration.py) with fields:
            'iteration', 'epoch', 'sample', 'time', 'test_now', 'done'
        scores is dict<(subset, prediction_function, cost_function) -> score>  where:
            subset is a string identifying the subset (eg 'train', 'test')
            prediction_function is identifies the prediction function (usually None, but can be used if you specify multiple prediction functions)
            cost_function is identifiers the cost function.
    """
    info_score_pairs = InfoScorePairSequence() if score_collection is None else score_collection
    for (x_mini, y_mini), info in zip_minibatch_iterate_info(dataset.training_set.xy, minibatch_size=minibatch_size, n_epochs=n_epochs, test_epochs=test_epochs):
        if info.test_now:
            rate = (info.time-last_time)/(info.epoch - last_epoch) if info.epoch>0 else float('nan')
            print 'Epoch {}.  Rate: {:.3g}s/epoch'.format(info.epoch, rate)
            last_epoch = info.epoch
            last_time = info.time
            test_pairs = _dataset_to_test_pair(dataset, include = test_on)
            score = assess_prediction_functions(test_pairs, functions=predict_fcn, costs=score_measure, print_results=True, prediction_minibatches=minibatch_size if predict_in_minibatches else None)
            p = InfoScorePair(info, score)
            info_score_pairs.append(p)
            # print p.get_table(remove_headers=len(info_score_pairs)>1)
            if test_callback is not None:
                test_callback(info, score)
        if not info.done:
            if pass_iteration_info_to_training:
                train_fcn(x_mini, y_mini, info)
            else:
                train_fcn(x_mini, y_mini)
            if training_callback is not None:
                training_callback(info, x_mini, y_mini)
    return info_score_pairs


def get_best_score(score_info_pairs, subset = 'test', prediction_function = None, score_measure = None, lower_is_better = False):
    """
    DEPRECATED!!!

    Given a list of (info, score) pairs which represet the progress over training, find the best score and return it.
    :param score_info_pairs: A list<(IterationInfo, dict)> of the type returned in train_and_test_online_predictor
    :param subset: 'train' or 'test' ... which subset to use to look for the best score
    :param prediction_function: Which prediction function (if there are multiple prediction functions, otherwise leave blank)
    :param score_measure: Which score measure (if there are multiple score measures, otherwise leave blank)
    :param lower_is_better: True if a lower score is better for the chosen score_measure
    :return: best_info, best_score
        best_info is an InterationInfo object
        best_score is a dict<(subset, prediction_function, score_measure) -> score>
    """
    assert len(score_info_pairs)>0, "You need to have at least one score to determine the best one."
    _, first_score = score_info_pairs[0]
    all_subsets, all_functions, all_measures = [remove_duplicates(s) for s in zip(*first_score.keys())]
    if prediction_function is None:
        assert len(all_functions)==1, "You did not specify prediction_function... options are: {}".format(all_functions)
        prediction_function = all_functions[0]
    if score_measure is None:
        assert len(all_measures)==1, "You did not specify a score_measure... options are: {}".format(all_measures)
        score_measure = all_measures[0]
    best_info = None
    best_score = None
    for info, score in score_info_pairs:
        this_is_the_best = best_score is None or \
            (score[subset, prediction_function, score_measure]<best_score[subset, prediction_function, score_measure]) == lower_is_better
        if this_is_the_best:
            best_score = score
            best_info = info
    return best_info, best_score


def print_best_score(score_info_pairs, **best_score_kwargs):
    # DEPRECATED!!!
    best_info, best_score = get_best_score(score_info_pairs, **best_score_kwargs)
    print_score_results(score=best_score, info=best_info)


class ParameterSchedule(object):

    def __init__(self, schedule, print_variable_name = None):
        """
        Given a schedule for a changing parameter (e.g. learning rate) get the values for this parameter at a given time.
        e.g.:
            learning_rate_scheduler = ScheduledParameter({0: 0.1, 10: 0.01, 100: 0.001}, print_variable_name='eta')
            new_learning_rate = learning_rate_scheduler.get_new_value(epoch=14)
            assert new_learning_rate == 0.01

        :param schedule: Can be:
            - A dict<epoch: value> where the epoch is a number indicating the training progress and the value
              indicates the value that the parameter should take.
            - A function which takes the epoch and returns a parameter value.
            - A number or array, in which case the value remains constant
        """
        if isinstance(schedule, (int, float, np.ndarray)):
            schedule = {0: schedule}
        if isinstance(schedule, dict):
            assert all(isinstance(num, (int, float)) for num in schedule.keys())
            self._reverse_sorted_schedule_checkpoints = sorted(schedule.keys(), reverse=True)
        else:
            assert callable(schedule)
        self.schedule = schedule
        self.print_variable_name = print_variable_name
        self.last_value = None  # Just used for print statements.

    def get_new_value(self, epoch):
        if isinstance(self.schedule, dict):
            new_value = self.schedule[(e for e in self._reverse_sorted_schedule_checkpoints if e <= epoch).next()]
        else:
            new_value = self.schedule(epoch)
        if self.last_value != new_value and self.print_variable_name is not None:
            print 'Epoch {}: {} = {}'.format(epoch, self.print_variable_name, new_value)
            self.last_value = new_value
        return new_value


forward_to_server = True


def should_I_forward_to_server():
    return forward_to_server


def set_forward_to_server(val=True):
    '''
    If set to false, makes sure that the process calling dbplot does not forward to a different plotting server.
    :param val:
    :return:
    '''
    global forward_to_server
    forward_to_server = val


from __future__ import print_function
import SocketServer
import logging
import socket
import struct
import sys
from artemis.config import get_artemis_config_value

ARTEMIS_LOGGER = logging.getLogger('artemis')


def get_socket(address,port):
    '''
    Returns a socket, bound to the next free port starting from the given port.
    :param port:
    :return: Tuple (socket, used_port)
    '''
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    while True:
        try:
            server_address = (address, port)
            sock.bind(server_address)
        except SocketServer.socket.error as exc:
            # ARTEMIS_LOGGER.info('Port', port, 'already in use')
            port += 1
            if exc.args[0] == 48 or exc.args[0] == 98:
                port += 1
            else:
                raise
        else:
            break
    return (sock,port)


def send_size(sock, data):
    try:
        sock.sendall(struct.pack('!I', len(data)))
        sock.sendall(data)
    except SocketServer.socket.error as exc:
        if exc.args[0] == 32:
            print("Broken pipe", file=sys.stderr)
            sys.exit(0)
        else:
            raise


def recv_bytes(sock, size):
    buf = b''
    while size:
        try:
            newbuf = sock.recv(size)
        except SocketServer.socket.error as exc:
            if exc.args[0] == 54:
                print("Connection reset by peer", file=sys.stderr)
                sys.exit(0)
            else:
                raise
        buf += newbuf
        size -= len(newbuf)
    return buf

def recv_size(sock):

    size_data = recv_bytes(sock, 4)
    size = struct.unpack('!I', size_data)[0]
    message = recv_bytes(sock,size)

    return message

def get_local_ips():
    '''
    Returns all the different ways that one can address the own machine.
    source: http://stackoverflow.com/a/274644/2068168
    :return: list of strings
    '''
    from netifaces import interfaces, ifaddresses, AF_INET

    ip_list = []
    for interface in interfaces():
        if AF_INET in ifaddresses(interface):
            for link in ifaddresses(interface)[AF_INET]:
                ip_list.append(link['addr'])
    return ip_list

def is_valid_ip(ip_address):
    '''
    checks if ip_address has a valid format
    :param address:
    :return:
    '''
    try:
        socket.inet_aton(ip_address)
        return True
    except:
        return False

def is_valid_port(port):
    '''
    checks if port is a valid string.
    This will make sure that any given string is numeric and between the range of 0 and 65535
    Source: http://stackoverflow.com/a/12968117/2068168
    :param port:
    :return:
    '''
    import re
    port_pattern = re.compile("^([0-9]{1,4}|[1-5][0-9]{4}|6[0-4][0-9]{3}|65[0-4][0-9]{2}|655[0-2][0-9]|6553[0-5])$")
    return port_pattern.match(port)


def get_remote_artemis_path(remote_ip):
    return get_artemis_config_value(
        section=remote_ip,
        option='artemis_path',
        default_generator=lambda: raw_input('Specify Remote Artemis Installation Path: '),
        write_default=True
        )



from contextlib import contextmanager
import pickle
from artemis.fileman.local_dir import make_file_dir, format_filename, get_local_path
from artemis.plotting.drawing_plots import redraw_figure
from artemis.plotting.manage_plotting import ShowContext
import os
__author__ = 'peter'
import logging
ARTEMIS_LOGGER = logging.getLogger('artemis')
logging.basicConfig()
from matplotlib import pyplot as plt

_supported_filetypes = ('.eps', '.jpeg', '.jpg', '.pdf', '.pgf', '.png', '.ps', '.raw', '.rgba', '.svg', '.svgz', '.tif', '.tiff', '.pkl')


def save_figure(fig, path, ext=None, default_ext = '.pdf'):
    """
    :param fig: The figure to show
    :param path: The absolute path to the figure.
    :param default_ext: The default extension to use, if none is specified.
    :return:
    """

    if ext is None:
        _, ext = os.path.splitext(path)
        if ext == '':
            path += default_ext
        else:
            assert ext in _supported_filetypes, "We inferred the extension '{}' from your filename, but it was not in the list of supported extensions: {}" \
                .format(ext, _supported_filetypes)
    else:
        path += ext if ext.startswith('.') else '.'+ext

    if '%L' in path:
        path = path.replace('%L', fig.get_label() if fig.get_label() is not '' else 'unnamed')
    path = format_filename(path)

    make_file_dir(path)
    if ext=='.pkl':
        with open(path, 'w') as f:
            pickle.dump(fig, f, protocol=pickle.HIGHEST_PROTOCOL)
    else:
        fig.savefig(path)

    ARTEMIS_LOGGER.info('Saved Figure: %s' % path)
    return path


@contextmanager
def interactive_matplotlib_context(on=True):
    old_mode = plt.isinteractive()
    plt.interactive(on)
    yield
    plt.interactive(old_mode)


def show_saved_figure(relative_loc, title=None):
    """
    Display a saved figure.

    Behaviour: this simply opens a window with the figure, and then continues executing the code.

    :param relative_loc: Relative path (within the data directory) to the figure.  Treated as an absolute path
        if it begins with "/"
    :return:
    """
    fig_path, ext = os.path.splitext(relative_loc)
    if title is None:
        _, title = os.path.split(fig_path)
    abs_loc = get_local_path(relative_loc)
    assert os.path.exists(abs_loc), '"%s" did not exist.  That is odd.' % (abs_loc, )
    if ext in ('.jpg', '.png', '.tif'):
        try:
            from PIL import Image
            Image.open(abs_loc).show()
        except ImportError:
            ARTEMIS_LOGGER.error("Cannot display image '%s', because PIL is not installed.  Go pip install pillow to use this.  Currently it is a soft requirement.")
    elif ext == '.pkl':
        with interactive_matplotlib_context():
            with open(abs_loc) as f:
                fig = pickle.load(f)
                fig.canvas.set_window_title(title)
    else:
        import webbrowser
        webbrowser.open('file://'+abs_loc)


class SaveFiguresOnShow(ShowContext):

    def __init__(self, path, also_show=True, default_ext = '.pdf'):
        """
        :param path: The path to the figure.  If it does not start with "/", it is assumed to be relative to the Data directory.
        :param also_show: Also show the figures.
        :param default_ext: The default extension to use, if none is specified.
        """
        self._path = path
        self._default_ext = default_ext
        self._locations = []
        ShowContext.__init__(self, self.save_figure, clear_others=not also_show)

    def save_figure(self, fig=None):
        if fig is None:
            fig = plt.gcf()
        loc = save_figure(fig, self._path, default_ext=self._default_ext)
        self._locations.append(loc)

    def get_figure_locs(self):
        return list(self._locations)

@contextmanager
def save_figures_on_close(dir, prefix='', default_ext = 'pdf', close_after =False):

    old_fignums = plt.get_fignums()

    yield

    new_figs = [fig for fig in plt.get_fignums() if fig ]

    for fig_no in new_figs:
        make_file_dir(path)
    fig.savefig(path)





import SocketServer
import os
import sys
import time
import signal
from artemis.plotting.plotting_backend import get_plotting_server_address
from artemis.remote.child_processes import check_ssh_connection, check_if_port_is_free, execute_command, ChildProcess, \
    ParamikoPrintThread, Nanny
from artemis.remote.utils import get_local_ips, get_socket, get_remote_artemis_path
from pytest import raises


ip_addresses = [get_plotting_server_address()]
if ip_addresses[0] not in get_local_ips():
    ip_addresses.append("127.0.0.1")


def test_check_ssh_connections():
    for ip_address in ip_addresses:
        if ip_address not in get_local_ips():
            check_ssh_connection(ip_address)


def test_check_if_port_is_free():
    for ip_address in ip_addresses:
        if ip_address not in get_local_ips():
            with raises(AssertionError):
                check_if_port_is_free(ip_address,80)
        else:
            sock, port = get_socket(ip_address,7005)
            # port is now not free:
            with raises(SocketServer.socket.error):
                check_if_port_is_free(ip_address, port)
            sock.close()
            # now no error to be expected
            check_if_port_is_free(ip_address,port)


def test_communications():
    for ip_address in ip_addresses:
        execute_command(ip_address=ip_address,
                    blocking=True,
                    command = "python -c 'from __future__ import print_function\nimport sys,time\nfor i in range(10): print(i, file=sys.stderr if i%2==0 else sys.stdout);sys.stdout.flush();time.sleep(0.3)'")


def test_kill_process_gently():

    for ip_address in ip_addresses:
        if ip_address in get_local_ips():
            command = "python %s"%os.path.join(os.path.dirname(__file__), "bogus_test_functions.py")
        else:
            remote_path = get_remote_artemis_path()
            command = "python %s"%os.path.join(remote_path, "remote/bogus_test_functions.py")

        cp = ChildProcess(ip_address, command)
        stdin , stdout, stderr = cp.execute_child_process()
        # stdin , stdout, stderr = ssh_conn.exec_command(command)

        pid=cp.get_pid()
        print("Pid: %s"%pid)
        #stdout
        t1 = ParamikoPrintThread(source_pipe=stdout, target_pipe=sys.stdout, prefix="stdout: ")
        t1.start()
        # stderr
        t2 = ParamikoPrintThread(source_pipe=stderr, target_pipe=sys.stderr, prefix="stderr: ")
        t2.start()

        print("Waiting 4 seconds")
        time.sleep(4)

        print ("Killing Process %s:" %(pid) )
        cp.kill()
        print("Waiting 4 seconds")
        time.sleep(4)
        if cp.is_alive():
            print("Process still alive")
        else:
            print("Process dead")


        print("Terminating")



def test_remote_graphics():
    for ip_address in ip_addresses:
        # command = 'export DISPLAY=:0.0; python -c "from matplotlib import pyplot as plt;import time; plt.figure();plt.show();time.sleep(10)"'
        if ip_address not in get_local_ips():
            command =["export DISPLAY=:0.0;", "python","-u", "-c", "from matplotlib import pyplot as plt;import time; plt.figure();plt.show();time.sleep(10)"]
        else:
            command =["python","-u", "-c", "from matplotlib import pyplot as plt;import time; plt.figure();plt.show();time.sleep(10)"]

        cp = ChildProcess(ip_address=ip_address,command=command)
        i, o, e = cp.execute_child_process()
        time.sleep(5)
        cp.deconstruct(message=signal.SIGTERM)
        print(o.read())
        print(e.read())


def test_is_alive():
    for ip in ip_addresses:
        command ="python -u -c 'from __future__ import print_function\nimport sys,time\nfor i in range(10): print(i, file=sys.stderr if i%2==0 else sys.stdout);time.sleep(0.3)'"
        cp = ChildProcess(ip_address=ip, command=command)
        i,o,e = cp.execute_child_process()
        t = ParamikoPrintThread(o,sys.stdout).start()
        t = ParamikoPrintThread(e,sys.stderr).start()
        while cp.is_alive():
            print("alive")
            time.sleep(0.3)
        print("dead")


if __name__ == "__main__":
    test_check_ssh_connections()
    test_check_if_port_is_free()
    test_communications()
    test_kill_process_gently()
    test_remote_graphics()
    test_is_alive()



from artemis.general.kwarg_dealer import KwargDealer
import pytest

__author__ = 'peter'


def my_favourite_system(planet, **kwargs):
    kd = KwargDealer(kwargs)

    if planet == 'Earth':
        params = kd.deal({'moon': 'The Moon', 'person': 'Jesus'})
        favsys = 'My favourite planet is %s, my favourite moon is %s, and my favourite person is %s' %(planet, params['moon'], params['person'])
    elif planet == 'Mars':
        params = kd.deal({'moon': 'Phobos'})
        favsys = 'My favourite planet is %s, my favourite moon is %s' % (planet, params['moon'])
    elif planet == 'Venus':
        favsys = 'My favourite planet is %s.' % (planet, )

    kd.assert_empty()
    return favsys


def test_kwarg_dealer():

    assert my_favourite_system('Earth') == 'My favourite planet is Earth, my favourite moon is The Moon, and my favourite person is Jesus'
    assert my_favourite_system('Earth', person = 'Margerit Thatcher') == 'My favourite planet is Earth, my favourite moon is The Moon, and my favourite person is Margerit Thatcher'
    assert my_favourite_system('Mars') == 'My favourite planet is Mars, my favourite moon is Phobos'
    assert my_favourite_system('Venus') == 'My favourite planet is Venus.'

    with pytest.raises(AssertionError):
        my_favourite_system('Mars', person = 'Jesus')  # No people on Mars

    with pytest.raises(AssertionError):
        my_favourite_system('Venus', moon = 'Europa')  # No moons around Venus

if __name__ == '__main__':
    test_kwarg_dealer()



from scipy.stats.stats import mode
import numpy as np

from artemis.ml.predictors.i_predictor import IPredictor


__author__ = 'peter'


class MockPredictor(IPredictor):

    def __init__(self, prediction_function):
        self._prediction_function = prediction_function

    def fit(self, input_data, target_data):
        pass

    def predict(self, input_data):
        return self._prediction_function(input_data)


class MostFrequentPredictor(IPredictor):

    def train(self, input_data, target_data):
        (self._most_frequent_value, ), _ = mode(target_data, axis = 0)
        self._target_type = target_data.dtype

    def predict(self, input_data):
        return np.zeros((len(input_data), )+self._most_frequent_value.shape, dtype = self._most_frequent_value.dtype)+self._most_frequent_value


class DistributionPredictor(IPredictor):

    def train(self, input_data, target_data):
        distro = np.sum(target_data, axis = 0).astype(float)
        self._distro = distro/np.sum(distro)

    def predict(self, input_data):
        return np.zeros((len(input_data), )+self._distro.shape, dtype = self._distro.dtype)+self._distro



import numpy as np
from matplotlib import pyplot as plt
from artemis.plotting.expanding_subplots import select_subplot


def test_expanding_subplots(block=False):

    select_subplot('agfdsfgdg')
    plt.plot(np.sin(np.linspace(0, 10, 100)))

    select_subplot('dsxfdsgf')
    plt.imshow(np.random.randn(10, 10))

    select_subplot('agfdsfgdg')
    plt.plot(np.cos(np.linspace(0, 10, 100)))

    plt.show(block = block)


def test_closing_fig(block=False):
    """
    Test that if we close the figure, we recognise that fact, and that we therefore need to recreate the subplot afterwards.
    """
    fig = plt.figure()

    select_subplot('fdsfsd')
    plt.plot(np.sin(np.linspace(0, 10, 100)))

    plt.close(fig)
    select_subplot('fdsfsd')
    plt.plot(np.cos(np.linspace(0, 10, 100)))

    plt.show(block = block)



if __name__ == '__main__':
    test_closing_fig()
    test_expanding_subplots()



import textwrap

from artemis.general.display import IndentPrint, CaptureStdOut, side_by_side, DocumentWrapper, deepstr
import numpy as np

_desired = """
aaa
bbb
    ccc
    ddd
        eee
        fff
    ggg
    hhh
iii
jjj
"""


def test_indent_print():

    with CaptureStdOut() as cap:
        print 'aaa'
        print 'bbb'
        with IndentPrint():
            print 'ccc'
            print 'ddd'
            with IndentPrint():
                print 'eee'
                print 'fff'
            print 'ggg'
            print 'hhh'
        print 'iii'
        print 'jjj'

    assert '\n'+cap.read() == _desired


str1 = '' + \
    'For those who are led by the Spirit of God are the\n' \
    'children of God.\n' \
    'Romans 8:14'

str2 = '' \
       'But to you who are listening I say: Love your enemies,\n' \
       'do good to those who hate you, bless those who curse you, pray\n' \
       'for those who mistreat you.\n' \
       'Luke 6:27-28'

desired = '' \
    'For those who are led by the Spirit of God are the    But to you who are listening I say: Love your enemies,        \n' \
    'children of God.                                      do good to those who hate you, bless those who curse you, pray\n' \
    'Romans 8:14                                           for those who mistreat you.                                   \n' \
    '                                                      Luke 6:27-28                                                  '


def test_side_by_side():

    print 'String 1:\n{}'.format(str1)
    print 'String 2:\n{}'.format(str2)

    out = side_by_side([str1, str2])
    print 'Side by side:\n{}'.format(out)
    assert out==desired  # Would work but pycharm automatically trims trailing spaces from the strings defined av


def test_document_wrapper():
    """
    Document Wrapper is deals with wrapping text with new lines already present.
    :return:
    """

    str3="0123456789\n0123456789\n01234567890123456789"

    desired3="0123456789\n0123456789\n012345678901\n23456789"

    w1 = textwrap.TextWrapper(width=12, replace_whitespace=False)
    r1 = w1.fill(str3)
    assert r1 != desired3
    assert r1 == "0123456789\n0123456789\n0\n123456789012\n3456789"

    w2 = DocumentWrapper(width=12, replace_whitespace=False)
    r2 = w2.fill(str3)
    assert r2 == desired3


def test_deepstr():

    obj = {'a': np.arange(100).reshape(10, 10), 'bbbb': [1, 3, np.arange(6).reshape(2, 3), ('xx', 'yy')]}
    obj['c'] = obj['bbbb']
    string_desc = deepstr(obj)
    print string_desc
    # For now, no assertions, because string contains IDS which will always change.  We can come up with some way to do this later with regular experessions if needed.


if __name__ == '__main__':
    test_indent_print()
    test_side_by_side()
    test_document_wrapper()
    test_deepstr()


import numpy as np

from artemis.ml.tools.processors import RunningAverage


__author__ = 'peter'


def test_running_average():

    inp = np.arange(5)
    processor = RunningAverage()
    out = [processor(el) for el in inp]
    assert out == [0, 0.5, 1, 1.5, 2]

    inp = np.random.randn(10, 5)
    processor = RunningAverage()
    out = [processor(el) for el in inp]
    assert all(np.allclose(out[i], np.mean(inp[:i+1], axis = 0)) for i in xrange(len(inp)))


if __name__ == '__main__':

    test_running_average()



import numbers
import numpy as np
from artemis.general.mymath import softmax
from artemis.general.numpy_helpers import get_rng
from artemis.general.should_be_builtins import bad_value

__author__ = 'peter'


"""
Some helper functions for working with neural networks.

For a more complete library including trainable neural nets, convnets, etc, see Plato.
https://github.com/petered/plato

Or, you know, a "mainstream" library, like Keras: https://keras.io/
"""


def initialize_network_params(layer_sizes, mag='xavier-both', base_dist='normal', include_biases = True, scale=1., rng=None):
    """
    Initialize parameters for a fully-connected neural network.

    :param layer_sizes: A list of integers indicating layer sizes (including that of the input layer)
    :param mag: The standard deviation, or a string identifying a method for selecting the standard deviation.
        String options can be:
            'xavier-forward' - Best for preserving variance of a linear, tanh, or sigmoidal network across layers.
            'xavier-both': - A compromize between preserving the variance of the forward, backward pass
            'xavier-relu': - Best for preserving variance on the forward pass in a ReLU net.
    :param base_dist: 'normal' or 'uniform', or a function taking (n_in, n_out) and returning a (n_in, n_out) array
    :param include_biases: Also create initial biases.
    :param rng: A random number generator or seed
    :return: A list of 2-tuples of (weight, bias) parameters (if include_biases is True) otherwise a list of weight matrices.

    Note: To get the weights/biases in separate lists, simply go:
        weights, biases = zip(*initialize_network_params(...))

    Note: See http://andyljones.tumblr.com/post/110998971763/an-explanation-of-xavier-initialization
    For a good explanation of the 'xavier' initialization schemes.
    """
    rng = get_rng(rng)
    ws = [initialize_weight_matrix(n_in, n_out, mag=mag, base_dist=base_dist, scale=scale, rng=rng) for n_in, n_out in zip(layer_sizes[:-1], layer_sizes[1:])]
    if include_biases:
        bs = [np.zeros(n_out) for n_out in layer_sizes[1:]]
        return zip(ws, bs)
    else:
        return ws


def initialize_weight_matrix(n_in, n_out, mag='xavier', base_dist='normal', scale=1., rng=None):
    """
    Initialize a weight matrix
    :param n_in: Number of input units
    :param n_out: Number of output units
    :param mag: The magnitude, or a string identifying how to calculate the magnitude.
        String options can be:
            'xavier-forward' - Best for preserving variance of a linear, tanh, or sigmoidal network across layers.
            'xavier-both': - A compromize between preserving the variance of the forward, backward pass
            'xavier-relu': - Best for preserving variance on the forward pass in a ReLU net.
    :param base_dist: 'normal' or 'uniform', or a function taking (n_in, n_out) and returning a (n_in, n_out) array
    :param rng: Random number generator or seed
    :return: A shape (n_in, n_out) initial weight matrix.
    """
    rng = get_rng(rng)

    w_base = rng.randn(n_in, n_out) if base_dist == 'normal' else \
        (np.rand(n_in, n_out) - 0.5)*np.sqrt(12) if base_dist=='uniform' else \
        bad_value(base_dist)

    mag_number = \
        np.sqrt(2./(n_in+n_out)) if mag in ('xavier', 'xavier-both') else \
        np.sqrt(1./n_in) if mag=='xavier-forward' else \
        np.sqrt(2./n_in) if mag=='xavier-relu' else \
        mag if isinstance(mag, numbers.Real) else \
        bad_value(mag)

    return w_base * (mag_number*scale)


def activation_function(data, function_name):
    if function_name=='relu':
        return np.maximum(0, data)
    elif function_name=='linear':
        return data
    elif function_name=='softmax':
        return softmax(data, axis=-1)
    elif function_name=='softplus':
        return np.log(np.exp(data)+1)
    elif function_name in ('sigm', 'sigmoid'):
        return 1./(1+np.exp(-data))
    elif function_name == 'tanh':
        return np.tanh(data)
    else:
        raise Exception('No Nonlinearity "{}".  Add it.'.format(function_name))


def forward_pass_activations(input_data, weights, biases = None, hidden_activations='relu', output_activations = 'relu'):
    """
    Return the activations from a forward pass of a ReLU net.
    :param input_data: A (n_frames, n_dims_in) array
    :param weights: A list of (n_dim_in, n_dim_out) weight matrices
    :param biases: An optional (len(weights)) list of (w.shape[1]) biases for each weight matrix
    :param hidden_activations: Indicates the hidden layer activation function
    :param output_activations: Indicates the output layer activation function
    :return: activations: A len(weights)*2+1 array, where:
        activations[::2] are the input and pre-nonlinearities (len(ws)+1)
        activations[1::2] are the post-nonlinerity activations (len(ws))
    """
    activations = [input_data]
    if biases is None:
        biases = [0]*len(weights)
    else:
        assert len(biases)==len(weights)
    x = input_data
    for i, (w, b) in enumerate(zip(weights, biases)):
        u = x.dot(w)+b
        x = activation_function(u, output_activations if i==len(weights)-1 else hidden_activations)
        activations += [u, x]
    return activations


def forward_pass(input_data, weights, biases = None, hidden_activations='relu', output_activation = 'relu'):
    """
    Do a forward pass of an MLP
    :param input_data: A (n_frames, n_dims_in) array
    :param weights: A list of (n_dim_in, n_dim_out) weight matrices
    :param biases: An optional (len(weights)) list of (w.shape[1]) biases for each weight matrix
    :param hidden_activations: Indicates the hidden layer activation function
    :param output_activation: Indicates the output layer activation function
    :return: output: A (input_data.shape[0], weights[-1].shape[1]) array of outputs
    """
    return forward_pass_activations(input_data=input_data, weights=weights, biases=biases, hidden_activations=hidden_activations, output_activations=output_activation)[-1]


from artemis.experiments.decorators import ExperimentFunction
from artemis.experiments.experiment_record_view import display_experiment_record, compare_experiment_results, \
    get_oneline_result_string, print_experiment_record_argtable, show_experiment_records
from artemis.experiments.experiments import experiment_testing_context
from artemis.general.display import CaptureStdOut


def display_it(result):
    print str(result) + 'aaa'


def one_liner(result):
    return str(result) + 'bbb'


def compare_them(results):
    print ', '.join('{}: {}'.format(k, results[k]) for k in sorted(results.keys()))


@ExperimentFunction(display_function=display_it, one_liner_function=one_liner, comparison_function=compare_them)
def my_xxxyyy_test_experiment(a=1, b=2):
    print 'xxx' if a==1 else 'yyy'
    return a+b


my_xxxyyy_test_experiment.add_variant('a2', a=2)


def test_experiments_function_additions():

    with experiment_testing_context():
        r1=my_xxxyyy_test_experiment.run()
        r2=my_xxxyyy_test_experiment.get_variant('a2').run()

        assert r1.get_log() == 'xxx\n'
        assert r2.get_log() == 'yyy\n'

        assert get_oneline_result_string(my_xxxyyy_test_experiment.get_latest_record()) == '3bbb'
        assert get_oneline_result_string(my_xxxyyy_test_experiment.get_variant('a2').get_latest_record()) == '4bbb'

        with CaptureStdOut() as cap:
            display_experiment_record(my_xxxyyy_test_experiment.get_latest_record())
        assert cap.read() == '3aaa\n'

        with CaptureStdOut() as cap:
            compare_experiment_results([my_xxxyyy_test_experiment, my_xxxyyy_test_experiment.get_variant('a2')])
        assert cap.read() == 'my_xxxyyy_test_experiment: 3, my_xxxyyy_test_experiment.a2: 4\n'

        print_experiment_record_argtable([r1, r2])

        show_experiment_records([r1, r2])


def test_experiment_function_ui():

    with experiment_testing_context():
        for existing_record in my_xxxyyy_test_experiment.get_variant_records(flat=True):
            existing_record.delete()
        assert len(my_xxxyyy_test_experiment.get_variant_records(flat=True)) == 0

        my_xxxyyy_test_experiment.browse(raise_display_errors=True, command='run all -p', close_after=True)
        assert len(my_xxxyyy_test_experiment.get_variant_records()) == 2

        my_xxxyyy_test_experiment.browse(raise_display_errors=True, command='argtable all', close_after=True)
        my_xxxyyy_test_experiment.browse(raise_display_errors=True, command='compare all', close_after=True)
        my_xxxyyy_test_experiment.browse(raise_display_errors=True, command='display all', close_after=True)
        my_xxxyyy_test_experiment.browse(raise_display_errors=True, command='show all', close_after=True)


if __name__ == '__main__':
    test_experiments_function_additions()
    test_experiment_function_ui()



import time


class ProgressIndicator(object):

    def __init__(self, expected_iterations, name=None, update_every = (2, 'seconds'), post_info_callback = None, just_use_last=False):

        self._expected_iterations = expected_iterations
        update_interval, update_unit = update_every
        assert update_unit in ('seconds', 'percent', 'iterations')

        if update_unit == 'percent':
            update_unit = 'iterations'
            update_interval = update_interval(expected_iterations)
        self.name = name
        self._update_unit = update_unit
        self._update_interval = update_interval
        self._start_time = time.time()
        self._should_update = {
            'iterations': self._should_update_iter,
            'seconds': self._should_update_time,
            }[self._update_unit]

        self._i = 0
        self._last_update = -float('inf')
        self._post_info_callback = post_info_callback
        self.just_use_last = just_use_last
        self._last_time = self._start_time
        self._last_progress = 0

    def __call__(self, iteration = None):
        self.print_update(iteration)

    def print_update(self, progress=None):
        self._current_time = time.time()
        if progress is None:
            progress = self._i
        if self._should_update() or progress == self._expected_iterations-1:
            elapsed = self._current_time - self._start_time
            frac = float(progress)/(self._expected_iterations-1)
            if self.just_use_last:
                remaining = (self._current_time - self._last_time)/(frac - self._last_progress) * (1-frac) if frac > 0 else float('NaN')
            else:
                remaining = elapsed * (1 / frac - 1) if frac > 0 else float('NaN')
            elapsed = self._current_time - self._start_time
            self._last_update = progress if self._update_unit == 'iterations' else self._current_time
            print 'Progress%s: %s%%.  %.1fs Elapsed, %.1fs Remaining.%s' \
                % ('' if self.name is None else ' of '+self.name, int(100*frac), elapsed, remaining, (', %s' % (self._post_info_callback(), )) if self._post_info_callback is not None else '')
        self._i += 1

        if self.just_use_last:
            self._last_time = self._current_time
            self._last_progress = frac

    def _should_update_time(self):
        return self._current_time-self._last_update > self._update_interval

    def _should_update_iter(self):
        return self._i - self._last_update > self._update_interval



try:
    from scipy import weave
except ImportError:
    print "Cannot Import scipy weave.  That's ok for now, you won't be able to use the fastplot function."

__author__ = 'peter'
import numpy as np
import matplotlib.pyplot as plt
"""
Functions to speed up pylab plotting, which can sometimes be unnecessairily slow.
"""


def fastplot(line_data, xscale = 'linear', yscale = 'linear', resolution = 2000, min_points = 20000):
    """
    Fast plot for those times when you have a lot of data points and it becomes too slow for pylab to handle.
    The plot produced here should look the same (unless you zoom) but display much faster.

    :param line_data: A vector of data
    :param xscale: {'linear', 'log'}
    :param yscale: {'linear', 'log'}
    :param resolution: The number intervals to bin points into
    :param min_points: The minimum number of points required to bother with this approach.
    :return: A plot handle
    """

    assert line_data.ndim == 1

    if len(line_data) < min_points:
        h= plt.plot(line_data)
    else:
        if xscale == 'linear':
            intervals = np.linspace(0, len(line_data), resolution)
        elif xscale == 'log':
            intervals = np.logspace(0, np.log10(len(line_data)), resolution)
        else:
            raise Exception("Can't yet deal with scale %s" % xscale)

        extreme_indices = find_interval_extremes(line_data, edges = intervals[1:])
        h=plt.plot(extreme_indices, line_data[extreme_indices])

    plt.gca().set_xscale(xscale)
    plt.gca().set_yscale(yscale)

    return h


def fastloglog(line_data, **kwargs):
    return fastplot(line_data, xscale='log', yscale = 'symlog', **kwargs)


def find_interval_extremes(array, edges):
    """
    Find the indeces of extreme points within each interval, and on the outsides of the two end edges.
    :param array: A vector
    :param edges: A vector of edges defining the intervals.  It's assumed that -Inf, Inf form the true outer edges.
    :return: A vector of ints indicating the indeces of extreme points.  If a distinct min and max extreme are found
        within every interval, this vector will have length 2*(len(edges)+1).  Otherwise, it will be shorter.
    """

    indices = np.zeros(len(array), dtype = int)-1
    how_many = np.array([0])
    code = """
    float min = INFINITY;
    float max = -INFINITY;
    int argmin;
    int argmax;
    bool foundpoint = false;
    int in_counter = 0;
    int out_counter = 0;
    int edge_counter = 0;
    while(in_counter<Narray[0]){
        float next_edge;
        if (edge_counter == Nedges[0])
            next_edge = INFINITY;
        else
            next_edge = edges[edge_counter];

        if (array[in_counter] < min){
            min = array[in_counter];
            argmin = in_counter;
            foundpoint = true;
        }
        if (array[in_counter] > max){
            max = array[in_counter];
            argmax = in_counter;
            foundpoint = true;
        }
        in_counter++;
        if (in_counter > next_edge){
            if (foundpoint){
                if (argmin < argmax){
                    indices[out_counter] = argmin;
                    indices[out_counter+1] = argmax;
                    out_counter+=2;
                }
                else if (argmax < argmin){
                    indices[out_counter] = argmax;
                    indices[out_counter+1] = argmin;
                    out_counter+=2;
                }
                else {
                    indices[out_counter] = argmax;
                    out_counter++;
                }
                min = INFINITY;
                max = -INFINITY;
                foundpoint = false;
            }
            edge_counter++;
        }
    }
    how_many[0] = out_counter;
    """
    weave.inline(code, ['array', 'edges', 'indices', 'how_many'], compiler = 'gcc')
    result = indices[:how_many[0]]
    return result



import numpy as np

from artemis.ml.predictors.predictor_comparison import assess_online_predictor
from artemis.ml.predictors.train_and_test import percent_argmax_correct
from plato.tools.common.bureaucracy import multichannel
from artemis.ml.datasets.synthetic_clusters import get_synthetic_clusters_dataset
from artemis.ml.tools.processors import OneHotEncoding


__author__ = 'peter'


def assert_online_predictor_not_broken(predictor_constructor, initial_score_under = 35, final_score_over = 95, n_epochs = 1,
        minibatch_size = 'full', categorical_target = False, accumulator = None, n_extra_tests = 0):
    """
    Assert that your predictor is not a total embarrassment.  (Note that it still may pass this test and be a terrible
    predictor, this at least makes clear that it's not completely broken.)

    :param predictor_constructor: A constructor that returns an IPredictor object given (n_dims_in, n_dims_out) as
        arguments.
    :param initial_score_under: Asser that the initial score on the 4-cluster dataset (where chance is 25%) is worse than
        this (mainly just makes sure you're not cheating somehow)
    :param final_score_over: Assert that the final score is over this - Solving this dataset isn't rocket science.  It
        is not hard to get a final score of 100.
    :param n_epochs: Now many epochs should you run?
    :param minibatch_size: Minibatch size.  By default, do full-batch training.
    :param categorical_target: If True, your predictor expects an integer as a target, where the integer indicates the
        correct label.  Otherwise, it expects a one-hot encoding vector - with the unit corresponding to the label being
        True.
    :param n_extra_tests: Number of extra tests - you may set this to non-zero to see the progress of your predictor
        over training.
    """
    dataset = get_synthetic_clusters_dataset(dtype = 'float32')

    if not categorical_target:
        dataset = dataset.process_with(targets_processor=multichannel(OneHotEncoding()))
        out_shape = dataset.target_size
    else:
        out_shape = dataset.n_categories

    predictor = predictor_constructor(dataset.input_size, out_shape)
    record = assess_online_predictor(predictor, dataset, evaluation_function=percent_argmax_correct,
        test_epochs=np.linspace(0, n_epochs, 2+n_extra_tests), minibatch_size=minibatch_size,
        accumulator = accumulator, test_on = 'test')
    scores = record.get_scores()
    assert scores[0] <= initial_score_under, "Initial score was %.2f%%, which was greater than expected (<%.2f%%).  That's odd." % (scores[0], initial_score_under)
    assert scores[-1] >= final_score_over, 'Achieved a final score of %.2f%%, which was less than the threshold of %.2f%%' % (scores[-1], final_score_over)



__author__ = 'peter'


class ReportCollector(object):

    def __init__(self, display = True):
        self.display = display
        self._report = []

    def append(self, update):
        self._report.append(update)
        if self.display:
            print update

    def get_report(self):
        return self._report

    def get_report_text(self):
        return '\n'.join(self._report)

    def print_report(self):
        print '='*15+' Report '+'='*15+'\n'+self.get_report_text()+'\n'+'='*40+'\n'



import logging
import sys


def execute_and_capture_locals(fcn, *args, **kwargs):
    """
    Execute the function with the provided arguments, and return both the output and
    the LOCAL VARIABLES of the function.  This is crazy.  It actually looks inside the
    function at the time that return is called and grabs all the local variables.
    """
    cap = CaptureLocals()

    with cap:
        out = fcn(*args, **kwargs)

    local_vars = cap.get_captured_locals()

    return out, local_vars


class CaptureLocals(object):
    """
    Use this object to capture the locals of a function that you call.
    """
    _profiler_stack = []

    def __init__(self):

        self._locals = None

        def tracer(frame, event, arg):
            if event == 'return':
                # Note - this is called for every return of every function called within.  The only
                # reason it's ok is that the final return is always that of the decorated function.
                # Still, we're often doing thousands of unnecessary copies.
                local_variables = frame.f_locals.copy()
                self._locals = local_variables

        self._profile_fcn = tracer

    def __enter__(self):
        CaptureLocals._profiler_stack.append(self._profile_fcn)
        sys.setprofile(self._profile_fcn)
        return self

    def __exit__(self, _, _1, _2):
        CaptureLocals._profiler_stack.pop()
        new_profiler = None if len(CaptureLocals._profiler_stack)==0 else CaptureLocals._profiler_stack[-1]
        sys.setprofile(new_profiler)

    def get_captured_locals(self):
        if self._locals is None:
            raise LocalsNotCapturedError()
        return self._locals


class LocalsNotCapturedError(Exception):

    def __init__(self):
        Exception.__init__(self, "Locals not captured.  This can happen if you're calling a function from the debugger or "
            "something else that messes with the profiler.")



import hashlib
import pickle
from collections import OrderedDict
import numpy as np


def fixed_hash_eq(obj1, obj2):
    """
    Return True if the objects have equal fixed-hashes.  You can use this instead of the "==" operator when comparing
    two objects that don't contain all python primatives.  All objects contained must be either primatives, numpy arrays,
    or extend the FixedHashObject interface.

    :param obj1: An arbitrary object
    :param obj2: Another arbitrary project,
    :return: A boolean, indicating whether their hashes are equal.
    """
    return compute_fixed_hash(obj1)==compute_fixed_hash(obj2)


def compute_fixed_hash(obj, hasher = None, try_objects=False):
    """
    Given an object, return a hash that will always be the same (not just for the lifetime of the
    object, but for all future runs of the program too).
    :param obj: Some nested container of primitives
    :param hasher: (for internal use - note that this is stateful, so calling this function with this argument changes
        the hasher object)
    :return: A 32-character hexidecimal (0-f) hash code for your object.
    """
    if hasher is None:
        hasher = hashlib.md5()
    hasher.update(obj.__class__.__name__)
    if isinstance(obj, np.ndarray):
        hasher.update(pickle.dumps(obj.dtype, protocol=2))
        hasher.update(pickle.dumps(obj.shape, protocol=2))
        hasher.update(obj.tostring())
    elif isinstance(obj, (int, str, float, bool)) or (obj is None) or (obj in (int, str, float, bool)):
        hasher.update(pickle.dumps(obj, protocol=2))
    elif isinstance(obj, (list, tuple)):
        hasher.update(str(len(obj)))  # Necessary to distinguish ([a, b], c) from ([a, b, c])
        for el in obj:
            compute_fixed_hash(el, hasher=hasher)
    elif isinstance(obj, dict):
        hasher.update(str(len(obj)))  # Necessary to distinguish ([a, b], c) from ([a, b, c])
        keys = obj.keys() if isinstance(obj, OrderedDict) else sorted(obj.keys())
        for k in keys:
            compute_fixed_hash(k, hasher=hasher)
            compute_fixed_hash(obj[k], hasher=hasher)
    elif isinstance(obj, FixedHashObject):  # See below... allows you to make custom hashables
        compute_fixed_hash(obj.get_hash_description(), hasher=hash)
    elif hasattr(obj, 'memo_hashable'):  # Deprecated, just here for back-compatibility
        compute_fixed_hash(obj.memo_hashable(), hasher=hasher)
    elif try_objects:
        raise NotImplementedError()
        # klass = obj.__class__.__name__
    else:
        # TODO: Consider whether to pickle by default.  Note that pickle strings are not necessairly the same for identical objects.
        raise NotImplementedError("Don't have a method for hashing this %s" % (obj, ))
    return hasher.hexdigest()


class FixedHashObject(object):
    """
    Implement this interface to create an object that you can create fixed hashes form.  You can then call
    compute_fixed_hash on this object.
    """

    def get_hash_description(self):
        """
        Override this method to return some primitive (python ond numpy only) data structure that represents your object.
        This will be used to create the fixed hash.  Be careful.  It's up to you to ensure that the data stryucture you
        return fully represents the state of your object.
        :return: Any data structure consisting only of Python and Numpy primatives.
        """
        raise NotImplementedError('You need to override this method to compute a fixed hash this object: {}'.format(self))



'''
This file is required to perform plotting server related tests. Its purpose is to be executed in a seperate process or on a remote machine, automatically.
'''



from __future__ import print_function
import sys
import atexit
import time
def function():
    i = 1
    while i<10:
        print(i, file=sys.stderr)
        sys.stderr.flush()
        time.sleep(1.0)
        i +=1


def atexit_function():
    print("atexit called")
    sys.exit(0)


if __name__ == "__main__":
    atexit.register(atexit_function)
    function()


import sys
import textwrap
from StringIO import StringIO
from contextlib import contextmanager

from artemis.fileman.local_dir import make_file_dir
from artemis.general.should_be_builtins import izip_equal
import numpy as np

__author__ = 'peter'


def arraystr(arr, print_threshold, summary_threshold):
    """
    :param arr: A string summary of an array.
    :param print_threshold:
    :param summary_threshold:
    :return:
    """
    if arr.size<print_threshold:
        return '<{type} with shape={shape}, dtype={dtype}, value={value}, at {id}>'.format(
            type=type(arr).__name__, shape=arr.shape, dtype=arr.dtype, value=str(arr).replace('\n', ','), id=hex(id(arr)))
    elif arr.size<summary_threshold:
        return '<{type} with shape={shape}, dtype={dtype}, in=[{min:.3g}, {max:.3g}], at {id}>'.format(
            type=type(arr).__name__, shape=arr.shape, dtype=arr.dtype, min = arr.min(), max=arr.max(), id=hex(id(arr)))
    else:
        return '<{type} with shape={shape}, dtype={dtype}, at {id}>'.format(
            type=type(arr).__name__, shape=arr.shape, dtype=arr.dtype, id=hex(id(arr)))


@contextmanager
def hold_numpy_printoptions(**kwargs):
    """
    Temporarily set the numpy print options.
    See https://docs.scipy.org/doc/numpy/reference/generated/numpy.set_printoptions.html
    :param kwargs: Print options (see link)
    """
    opts = np.get_printoptions()
    np.set_printoptions(**kwargs)
    yield
    np.set_printoptions(**opts)


def str_with_arrayopts(obj, float_format='.3g', threshold=8, **kwargs):
    """
    Print
    :param obj:
    :param float_format:
    :param threshold:
    :param kwargs:
    :return:
    """
    with hold_numpy_printoptions(formatter = {'float': float_format}, threshold=threshold, **kwargs):
        return str(obj)


def deepstr(obj, memo=None, array_print_threhold = 8, array_summary_threshold=10000, indent ='  ', float_format = ''):
    """
    A recursive, readable print of a data structure.
    """
    if memo is None:
        memo = set()

    if id(obj) in memo:
        return "<{type} at {loc}> already listed".format(type=type(obj).__name__, loc=hex(id(obj)))
    memo.add(id(obj))

    if isinstance(obj, np.ndarray):
        string_desc = arraystr(obj, print_threshold=array_print_threhold, summary_threshold=array_summary_threshold)
    elif isinstance(obj, (list, tuple, set, dict)):
        kwargs = dict(memo=memo, array_print_threhold=array_print_threhold, array_summary_threshold=array_summary_threshold, indent=indent, float_format=float_format)

        if isinstance(obj, (list, tuple)):
            keys, values = [str(i) for i in xrange(len(obj))], obj
        elif isinstance(obj, dict):
            keys, values = obj.keys(), obj.values()
        elif isinstance(obj, set):
            keys, values = ['- ']*len(obj), obj
        else:
            raise Exception('Should never be here')

        max_indent = max(len(k) for k in keys)

        elements = ['{k}: {v}'.format(k=k, v=' '*(max_indent-len(k)) + indent_string(deepstr(v, **kwargs), indent=' '*max_indent, include_first=False)) for k, v in izip_equal(keys, values)]
        string_desc = '<{type} at {id}>\n'.format(type = type(obj).__name__, id=hex(id(obj))) + indent_string('\n'.join(elements), indent=indent)
        return string_desc
    elif isinstance(obj, float):
        string_desc = '{{:{}}}'.format(float_format).format(obj)
    else:
        string_desc = str(obj)
    return string_desc


_ORIGINAL_STDOUT = sys.stdout
_ORIGINAL_STDERR = sys.stderr


class CaptureStdOut(object):
    """
    An logger that both prints to stdout and writes to file.
    """

    def __init__(self, log_file_path = None, print_to_console = True):
        """
        :param log_file_path: The path to save the records, or None if you just want to keep it in memory
        :param print_to_console:
        """
        self._print_to_console = print_to_console
        if log_file_path is not None:
            # self._log_file_path = os.path.join(base_dir, log_file_path.replace('%T', now))
            make_file_dir(log_file_path)
            self.log = open(log_file_path, 'w')
        else:
            self.log = StringIO()
        self._log_file_path = log_file_path
        self.terminal = _ORIGINAL_STDOUT

    def __enter__(self):
        sys.stdout = self
        sys.stderr = self
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout = _ORIGINAL_STDOUT
        sys.stderr = _ORIGINAL_STDERR
        self.close()

    def get_log_file_path(self):
        assert self._log_file_path is not None, "You never specified a path when you created this logger, so don't come back and ask for one now"
        return self._log_file_path

    def write(self, message):
        if self._print_to_console:
            self.terminal.write(message)
        self.log.write(message)
        self.log.flush()

    def close(self):
        if self._log_file_path is not None:
            self.log.close()

    def read(self):
        if self._log_file_path is None:
            return self.log.getvalue()
        else:
            with open(self._log_file_path) as f:
                txt = f.read()
            return txt

    def __getattr__(self, item):
        return getattr(self.terminal, item)


def indent_string(str, indent = '  ', include_first = True):
    if include_first:
        return indent + str.replace('\n', '\n'+indent)
    else:
        return str.replace('\n', '\n'+indent)


class IndentPrint(object):
    """
    Indent all print statements
    """

    def __init__(self, block_header=None, spacing = 4, show_line = False, show_end = False):
        self.indent = '|'+' '*(spacing-1) if show_line else ' '*spacing
        self.show_end = show_end
        self.block_header = block_header

    def __enter__(self):
        self.old_stdout = sys.stdout
        if self.block_header is not None:
            print self.block_header
        sys.stdout = self

    def flush(self):
        self.old_stdout.flush()

    def write(self, message):
        if message=='\n':
            new_message = '\n'
        else:
            new_message = indent_string(message, self.indent)
        self.old_stdout.write(new_message)

    def close(self):
        self.old_stdout.close()

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout = self.old_stdout
        if self.show_end:
            print '```'


class DocumentWrapper(textwrap.TextWrapper):

    def wrap(self, text):
        split_text = text.split('\n')
        lines = [line for para in split_text for line in textwrap.TextWrapper.wrap(self, para)]
        return lines


def side_by_side(multiline_strings, gap=4, max_linewidth=None):
    """
    Return a string that displays two multiline strings side-by-side.
    :param multiline_strings: A list of multi-line strings (ie strings with newlines)
    :param gap: The gap (in spaces) to make between side-by-side display of the strings
    :param max_linewidth: Maximum width of the strings
    :return: A single string which shows all the above strings side-by-side.
    """
    if max_linewidth is not None:
        w = DocumentWrapper(width=max_linewidth, replace_whitespace=False)
        lineses = [w.wrap(mlstring) for mlstring in multiline_strings]
    else:
        lineses = [s.split('\n') for s in multiline_strings]
    longests = [max(len(line) for line in lines) for lines in lineses]

    spacer = ' '*gap
    new_lines = []
    for i in xrange(max(len(lines) for lines in lineses)):
        line = [lines[i] if i<len(lines) else '' for lines in lineses]
        final_line = [line + ' '*(max_length-len(line)) for line, max_length in zip(line, longests)]
        new_lines.append(spacer.join(final_line))
    return '\n'.join(new_lines)


def truncate_string(string, truncation, message = ''):
    """
    Truncate a string to a given length.  Optionally add a message at the end
    explaining the truncation
    :param string: A string
    :param truncation: An int
    :param message: A message, e.g. '...<truncated>'
    :return: A new string no longer than truncation
    """
    if truncation is None:
        return string
    assert isinstance(truncation, int)
    if len(string)>truncation:
        return string[:truncation-len(message)]+message
    else:
        return string



from abc import ABCMeta, abstractmethod
from itertools import cycle

from artemis.general.should_be_builtins import bad_value
from artemis.plotting.data_conversion import put_data_in_grid, RecordBuffer, data_to_image, put_list_of_images_in_array, \
    UnlimitedRecordBuffer
from matplotlib import pyplot as plt
import numpy as np

__author__ = 'peter'


class IPlot(object):

    __metaclass__ = ABCMeta

    @abstractmethod
    def update(self):
        pass

    @abstractmethod
    def plot(self):
        pass


class HistoryFreePlot(IPlot):

    def update(self, data):
        self._last_full_plot_data = data

    def plot(self):
        self._plot_last_data(self._last_full_plot_data)

    @abstractmethod
    def _plot_last_data(self, data):
        pass


colour_cycle = 'bgrmcyk'


class ImagePlot(HistoryFreePlot):

    def __init__(self, interpolation = 'nearest', show_axes = False, show_clims = True, clims = None, only_grow_clims = False, aspect = 'auto', cmap = 'gray', is_colour_data = None):
        """
        :param interpolation: How to interpolate array to form the image {'none', 'nearest', ''bilinear', 'bicubic', ... (see plt.imshow)}
        :param show_axes: Show axes marks (numbers along the side showing pixel locations)
        :param show_clims: Print the range of the colour scale at the bottom
        :param clims: (lower, upper) limit to colour-scale (or None to set it to the range of the data)
        :param aspect: 'auto' to stretch the image to the shape of the plot, 'equal' to maintain aspect ratio
        :param cmap: Colourmap {'gray', 'jet', ...}
        :param is_colour_data: Identify whether the image consists of colour data.  Usually, if you leave this at None,
            it will figure out the correct thing automatically, but this can, for instance, be used to distinquish a case
            where an image of shape shape is (12, 18, 3) should be interpreted as a (12x18) colour image or 12 18x3
            black and white images.  (Default in this case would be colour image)
        """
        self._plot = None
        self._interpolation = interpolation
        self._show_axes = show_axes
        self._clims = clims
        self._aspect = aspect
        self._cmap = cmap
        self._is_colour_data = is_colour_data
        self.show_clims = show_clims
        self.only_grow_clims = only_grow_clims
        if only_grow_clims:
            self._old_clims = (float('inf'), -float('inf'))

    def _plot_last_data(self, data):

        if len(data)==0:
            plottable_data = np.zeros((16, 16, 3), dtype = np.uint8)
            clims = (0, 1)
        else:
            if isinstance(data, list):
                data = put_list_of_images_in_array(data, fill_colour=(np.nan, np.nan, np.nan))
            elif data.ndim == 1:
                data = data[None]

            clims = ((np.nanmin(data), np.nanmax(data)) if data.size != 0 else (0, 1)) if self._clims is None else self._clims

            if self.only_grow_clims:
                clims = min(self._old_clims[0], clims[0]), max(self._old_clims[1], clims[1])
                self._old_clims = clims

            if self._is_colour_data is None:
                self._is_colour_data = data.shape[-1]==3

            plottable_data = put_data_in_grid(data, clims = clims, cmap = self._cmap, is_color_data = self._is_colour_data, fill_colour = np.array((0, 0, 128)), nan_colour = np.array((0, 0, 128))) \
                if not (self._is_colour_data and data.ndim==3 or data.ndim==2) else \
                data_to_image(data, clims = clims, cmap = self._cmap, nan_colour = np.array((0, 0, 128)))

        if self._plot is None:
            self._plot = plt.imshow(plottable_data, interpolation = self._interpolation, aspect = self._aspect, cmap = self._cmap)
            if not self._show_axes:
                self._plot.axes.tick_params(labelbottom = 'off')
                self._plot.axes.get_yaxis().set_visible(False)
        else:
            self._plot.set_array(plottable_data)
        if self.show_clims:
            self._plot.axes.set_xlabel('{:.3g} <> {:.3g}'.format(*clims))


class MovingImagePlot(ImagePlot):

    def __init__(self, buffer_len = 100, **kwargs):
        ImagePlot.__init__(self, **kwargs)
        self._buffer = RecordBuffer(buffer_len)

    def update(self, data):
        if np.isscalar(data):
            data = np.array([data])
        elif data.ndim != 1 and data.size == np.max(data.shape):
            data = data.flatten()
        else:
            assert data.ndim == 1
        buffer_data = self._buffer(data)
        ImagePlot.update(self, buffer_data.T)


class LinePlot(HistoryFreePlot):

    def __init__(self, x_axis_type = 'lin', y_axis_type = 'lin', x_bounds = (None, None), y_bounds = (None, None), y_bound_extend = (.05, .05),
                 x_bound_extend = (0, 0), make_legend = None, axes_update_mode = 'fit', add_end_markers = False, legend_entries = None,
                 legend_entry_size = 8, plot_kwargs = {}, allow_axis_offset = False):
        """
        :param y_axis_type: 'lin' (only one supported now)
        :param x_bounds: A tuple of (lower_bound, upper_bound), where None means automatic
        :param y_bounds: A tuple of (lower_bound, upper_bound), where None means automatic
        :param y_bound_extend: A tuple of (float, float) indicating how much to extend the automatically determined bounds.
        :param x_bound_extend: A tuple of (float, float) indicating how much to extend the automatically determined bounds.
        :param make_legend: True to make a legend
            'fit': Fit to current curve
            'expand': Expand axes to contain curve (useful if plotting over pre-existing curve)
        :param add_end_markers: Add a circle to indicate a start, and an "X" to indicate the end of each line.
        :param legend_entries: Entries of the legend.  Should be one per line in the data.
        :param legend_entry_size: Font size for legend entries.
        :param plot_kwargs:
        """
        # assert y_axis_type == 'lin', 'Changing axis scaling not supported yet'
        self.x_axis_type = x_axis_type
        self.y_axis_type = y_axis_type
        self._plots = None
        self.x_bounds = x_bounds
        self.y_bounds = y_bounds
        self.x_bound_extend = x_bound_extend
        self.y_bound_extend = y_bound_extend
        self.make_legend=make_legend
        self.plot_kwargs = plot_kwargs
        self.axes_update_mode = axes_update_mode
        self.add_end_markers = add_end_markers
        self._end_markers = []
        self.legend_entries = [legend_entries] if isinstance(legend_entries, basestring) else legend_entries
        self.legend_entry_size = legend_entry_size
        self.allow_axis_offset = allow_axis_offset

    def _plot_last_data(self, data):
        """
        :param data: Can be:
            An array of y_data (x_data is assumed to be np.arange(data.shape[0])
            A 2-tuple of (x_data, y_data)
        """

        # Format all data as (list<x_data>, list<y_data>)
        if isinstance(data, tuple) and len(data)==2:
            x_data, y_data = data
        else:
            x_data = None
            y_data = data
        if isinstance(y_data, np.ndarray):
            n_lines = 1 if y_data.ndim==1 else y_data.shape[1]
            x_data = [np.arange(1, y_data.shape[0]+1) if self.x_axis_type=='log' else np.arange(y_data.shape[0])] * n_lines if x_data is None else [x_data] * n_lines if x_data.ndim==1 else x_data.T
            # x_data = y_data.T if y_data.ndim==2 else y_data[None] if y_data.ndim==1 else bad_value(y_data.ndim)
            y_data = y_data.T if y_data.ndim==2 else y_data[None] if y_data.ndim==1 else bad_value(y_data.ndim, "Line plot data must be 1D or 2D, not {}D".format(y_data.ndim))
        else:  # List of arrays
            if all(d.ndim==0 for d in data):  # Turn it into one line
                x_data = np.arange(1, y_data.shape[0]+1) if self.x_axis_type=='log' else np.arange(y_data.shape[0])[None, :]
                y_data = np.array(data)[None, :]
            else:  # List of arrays becomes a list of lines
                x_data = [np.arange(1, y_data.shape[0]+1) if self.x_axis_type=='log' else np.arange(y_data.shape[0]) for d in y_data] if x_data is None else x_data
        assert len(x_data)==len(y_data), "The number of lines in your x-data (%s) does not match the number in your y-data (%s)" % (len(x_data), len(y_data))

        lower, upper = (np.nanmin(y_data) if self.y_bounds[0] is None else self.y_bounds[0], np.nanmax(y_data)+1e-9 if self.y_bounds[1] is None else self.y_bounds[1])
        left, right = (np.nanmin(x_data) if self.x_bounds[0] is None else self.x_bounds[0], np.nanmax(x_data)+1e-9 if self.x_bounds[1] is None else self.x_bounds[1])

        if left==right:
            right+=1e-9

        # Expand x_bound:
        delta = right-left if left-right >0 else 1e-9
        left -= self.x_bound_extend[0]*delta
        right += self.x_bound_extend[1]*delta

        # Expand y_bound:
        if self.y_axis_type=='lin':
            delta = upper-lower if upper-lower >0 else 1e-9
            lower -= self.y_bound_extend[0]*delta
            upper += self.y_bound_extend[1]*delta
        elif self.y_axis_type=='log':
            lower *= (1-self.y_bound_extend[0])
            upper *= (1+self.y_bound_extend[1])
        else:
            raise Exception(self.y_axis_type)

        if self._plots is None:
            self._plots = []
            plt.gca().autoscale(enable=False)
            plt.gca().get_yaxis().get_major_formatter().set_useOffset(self.allow_axis_offset)
            plt.gca().get_xaxis().get_major_formatter().set_useOffset(self.allow_axis_offset)
            if self.x_axis_type!='lin':
                plt.gca().set_xscale(self.x_axis_type)
            if self.y_axis_type!='lin':
                plt.gca().set_yscale(self.y_axis_type)

            if isinstance(self.plot_kwargs, dict):
                plot_kwargs = [self.plot_kwargs]*len(x_data)
            elif isinstance(self.plot_kwargs, (list, tuple)):
                assert len(self.plot_kwargs)==len(x_data), "You provided a list of {0} plot kwargs, but {1} lines".format(len(self.plot_kwargs), len(x_data))
                plot_kwargs = self.plot_kwargs
            for i, (xd, yd, legend_entry) in enumerate(zip(x_data, y_data, self.legend_entries if self.legend_entries is not None else [None]*len(x_data))):
                p, =plt.plot(xd, yd, label = legend_entry, **plot_kwargs[i])
                self._plots.append(p)
                _update_axes_bound(p.axes, (left, right), (lower, upper), self.axes_update_mode)
                if self.add_end_markers:
                    colour = p.get_color()
                    self._end_markers.append((plt.plot(xd[[0]], yd[[0]], marker='.', markersize=20, color=colour)[0], plt.plot(xd[0], yd[0], marker='x', markersize=10, mew=4, color=colour)[0]))

            if (self.make_legend is True) or (self.make_legend is None and (self.legend_entries is not None or len(y_data)>1)):
                plt.legend(loc='best', framealpha=0.5, prop={'size':self.legend_entry_size})
                # entries = [str(i) for i in xrange(len(y_data))] if self.legend_entries is None else self.legend_entries
                # assert len(self._plots) == len(self.legend_entries), 'You have %s plots but you specified %s entries for the legend: %s' % (len(self._plots), len(entries), entries)
                # handles, labels = plt.gca().get_legend_handles_labels()
                # if len(handles)==0:
                #     plt.legend(handles + self._plots, labels+entries, loc='best', prop={'size':8})
                # else:
                #     plt.gca().set_legend_handles_labels(handles + self._plots, labels+entries)

        else:
            for i, (p, xd, yd) in enumerate(zip(self._plots, x_data, y_data)):
                p.set_xdata(xd)
                p.set_ydata(yd)
                _update_axes_bound(p.axes, (left, right), (lower, upper), self.axes_update_mode)
                if self.add_end_markers:
                    self._end_markers[i][0].set_xdata(xd[[0]])
                    self._end_markers[i][0].set_ydata(yd[[0]])
                    self._end_markers[i][1].set_xdata(xd[[-1]])
                    self._end_markers[i][1].set_ydata(yd[[-1]])

        # plt.legend(loc='best', framealpha=0.5, prop={'size': self.legend_entry_size})


def _update_axes_bound(ax, (left, right), (lower, upper), mode = 'fit'):
    if mode=='fit':
        ax.set_xbound(left, right)
        ax.set_ybound(lower, upper)
    elif mode=='expand':
        old_left, old_right = ax.get_xbound()
        old_lower, old_upper = ax.get_ybound()
        if (old_left, old_right, old_lower, old_upper) == (0, 1, 0, 1):  # Virgin axes (probably)... overwrite them
            ax.set_xbound(left, right)
            ax.set_ybound(lower, upper)
        else:
            ax.set_xbound(min(old_left, left), max(old_right, right))
            ax.set_ybound(min(old_lower, lower), max(old_upper, upper))
    else:
        raise Exception('No axis update mode: "%s"' % (mode, ))


class MovingPointPlot(LinePlot):

    def __init__(self, buffer_len=None, **kwargs):
        """
        :param buffer_len: An integar to keep a fixed-length window, or None to keep an expanding buffer
        :param kwargs:
        :return:
        """
        LinePlot.__init__(self, **kwargs)
        self._buffer = UnlimitedRecordBuffer() if buffer_len is None else RecordBuffer(buffer_len)
        self.x_data = np.arange(-buffer_len+1, 1) if buffer_len is not None else None

    def update(self, data):
        if not np.isscalar(data) or isinstance(data, np.ndarray):
            data = np.array(data)
        buffer_data = self._buffer(data)
        x_data = np.arange(len(buffer_data)) if self.x_data is None else self.x_data
        LinePlot.update(self, (x_data, buffer_data))

    def plot(self):
        LinePlot.plot(self)


class Moving2DPointPlot(LinePlot):

    def __init__(self, buffer_len=None, **kwargs):
        LinePlot.__init__(self, add_end_markers=True, **kwargs)
        self._y_buffer = UnlimitedRecordBuffer() if buffer_len is None else RecordBuffer(buffer_len)
        self._x_buffer = UnlimitedRecordBuffer() if buffer_len is None else RecordBuffer(buffer_len)

    def update(self, (x_data, y_data)):

        x_buffer_data = self._x_buffer(x_data)
        y_buffer_data = self._y_buffer(y_data)

        valid_sample_start = np.argmax(~np.any(np.isnan(y_buffer_data.reshape(y_buffer_data.shape[0], -1)), axis=1))
        LinePlot.update(self, (x_buffer_data[valid_sample_start:], y_buffer_data[valid_sample_start:]))

    def plot(self):
        LinePlot.plot(self)


class TextPlot(IPlot):

    def __init__(self, max_history = 8, horizontal_alignment = 'left', vertical_alignment = 'bottom', size = 'medium'):
        """
        :param horizontal_alignment: {'left', 'center', 'right'}
        :param vertical_alignment: {'top', 'center', 'bottom', 'baseline'}
        :param size: [size in points | "xx-small" | "x-small" | "small" | "medium" | "large" | "x-large" | "xx-large" ]
        :return:
        """
        self._buffer = RecordBuffer(buffer_len = max_history, initial_value='')
        self._max_history = 10
        self._text_plot = None
        self.horizontal_alignment = horizontal_alignment
        self.vertical_alignment = vertical_alignment
        self.size = size
        self._x_offset = {'left': 0.05, 'center': 0.5, 'right': 0.95}[self.horizontal_alignment]
        self._y_offset = {'bottom': 0.05, 'center': 0.5, 'top': 0.95}[self.vertical_alignment]

    def update(self, string):
        if not isinstance(string, basestring):
            string = str(string)
        history = self._buffer(string)
        self._full_text = '\n'.join(history)

    def plot(self):
        if self._text_plot is None:
            ax = plt.gca()
            ax.get_xaxis().set_visible(False)
            ax.get_yaxis().set_visible(False)
            ax.set_axis_off()
            self._text_plot = ax.text(self._x_offset, self._y_offset, self._full_text, horizontalalignment=self.horizontal_alignment, verticalalignment=self.vertical_alignment, size = self.size)
        else:
            self._text_plot.set_text(self._full_text)


class BarPlot(HistoryFreePlot):

    def __init__(self, fill_width=0.8, ):
        self._plots = None
        self.fill_width = fill_width

    def _plot_last_data(self, data):
        """
        :param data: Can be:
            A vector of values
            A (n_values, n_measures) array of values
        """
        if data.ndim==1:
            data = data[:, None]

        if self._plots is None:
            n_values, n_sources = data.shape
            width = self.fill_width/n_sources
            x_centers = np.arange(n_values)
            self._plots = [plt.bar(x_centers+width*i-self.fill_width/2., data[:, i], width=width, color=c, edgecolor=c) for i, c in zip(xrange(n_sources), cycle(colour_cycle))]
            plt.xticks(x_centers)
        else:
            for i, plot in enumerate(self._plots):
                for rect, h in zip(plot, data[:, i]):
                    rect.set_height(h)
            _update_axes_bound(rect.axes, (-.5, data.shape[0]-.5), (min(0, np.min(data)), max(0, np.max(data))), mode='fit')


class HistogramPlot(IPlot):

    def __init__(self, edges, mode = 'mass', plot_type = 'bar', cumulative = False):
        assert mode in ('mass', 'density')
        edges = np.array(edges)
        self._edges = edges
        self._mode = mode
        self._binvals = np.ones(len(edges)-1)/len(edges)
        self._n_points = 0
        self._plot = None
        self._widths = np.diff(edges)
        self._lefts = edges[:-1]
        self._plot_type=plot_type
        self._cumulative = cumulative

    def update(self, data):

        if isinstance(data, (list, tuple)):
            data = np.array(data)

        # Update data
        new_n_points = self._n_points + data.size
        this_hist, _ = np.histogram(data, self._edges)
        frac = (float(data.size)/self._n_points) if self._n_points > 0 else 1
        self._binvals += this_hist * frac
        self._binvals /= max(1, np.sum(self._binvals))
        self._n_points = new_n_points
        # DIsplay
        heights = self._binvals if self._mode == 'mass' else self._binvals/self._widths
        if self._cumulative:
            heights = np.cumsum(heights)
        self._last_heights = heights

    def plot(self):
        heights = self._last_heights
        if self._plot_type == 'bar':
            if self._plot is None:
                self._plot = plt.bar(self._lefts, heights, width = self._widths)
            else:
                for rect, h in zip(self._plot, heights):
                    rect.set_height(h)
        elif self._plot_type == 'line':
            if self._plot is None:
                self._plot = plt.plot(self._edges[:-1], heights)
            else:
                self._plot[0].set_ydata(heights)

        self._plot[0].axes.set_ybound(0, np.max(heights)*1.05)


class CumulativeLineHistogram(HistogramPlot):

    def __init__(self, edges):
        HistogramPlot.__init__(self, edges, mode = 'mass', plot_type='line', cumulative=True)


def get_plot_from_data(data, mode, **plot_preference_kwargs):

    return \
        get_live_plot_from_data(data, **plot_preference_kwargs) if mode == 'live' else \
        get_static_plot_from_data(data, **plot_preference_kwargs) if mode == 'static' else \
        ImagePlot(**plot_preference_kwargs) if mode == 'image' else \
        bad_value(mode, 'Unknown plot modee: %s' % (mode, ))


def get_live_plot_from_data(data, line_to_image_threshold = 8, cmap = 'gray'):

    # TODO: Maybe refactor that so that plot objects contain their own "data validation" code, and we can
    # simply ask plots in sequence whether they can handle the data.
    if isinstance(data, basestring):
        return TextPlot()

    if isinstance(data, list):
        if all(isinstance(d, np.ndarray) for d in data):
            if all(d.ndim in (2, 3) for d in data):
                return ImagePlot()
            elif all(d.ndim==1 for d in data):
                return LinePlot()
            elif all(d.ndim==0 for d in data):
                return LinePlot()
            else:
                raise Exception("Don't know how to deal with a list of arrays with shapes: %s" % ([d.shape for d in data], ))
        elif all(np.isscalar(d) for d in data):
            return MovingPointPlot()
    elif isinstance(data, tuple):
        if len(data)==2:
            if np.isscalar(data[0]) and np.isscalar(data[1]):
                return Moving2DPointPlot()
            else:
                return LinePlot()
        else:
            raise Exception("Don't know what to do with a length-%s tuple of data" % (len(data)))

    is_scalar = np.isscalar(data) or data.shape == ()
    if is_scalar:
        data = np.array(data)

    is_1d = not is_scalar and data.size == np.max(data.shape)
    few_values = data.size < line_to_image_threshold

    if is_scalar or is_1d and few_values:
        return MovingPointPlot()
    elif is_1d:
        return MovingImagePlot()
    elif data.ndim == 2 and data.shape[1]<line_to_image_threshold:
        return LinePlot()
    elif data.ndim in (2, 3, 4, 5):
        return ImagePlot(cmap=cmap)
    else:
        raise NotImplementedError('We have no way to plot data of shape %s.  Make one!' % (data.shape, ))


def get_static_plot_from_data(data, line_to_image_threshold=8, cmap = 'gray'):

    if isinstance(data, basestring):
        return TextPlot()

    is_scalar = np.isscalar(data) or data.shape == ()
    if is_scalar or data.size==1:
        return TextPlot()

    is_1d = not is_scalar and data.size == np.max(data.shape)
    if is_1d:
        n_unique = len(np.unique(data))
        if n_unique == 2:
            return ImagePlot(cmap=cmap)
        else:
            return LinePlot()
    elif data.ndim == 2 and data.shape[1] < line_to_image_threshold:
        return LinePlot()
    else:
        return ImagePlot(cmap=cmap)



import numpy as np
from artemis.experiments import experiment_function
from artemis.general.mymath import softmax
from artemis.ml.datasets.mnist import get_mnist_dataset
from artemis.ml.predictors.logistic_regressor import onehot
from artemis.ml.tools.iteration import minibatch_index_info_generator
from matplotlib import pyplot as plt

"""
This demo shows how you can use the Artemis Experiment Framework.

Using the Experiment framework (via the UI) requires only 2 additional lines of code:
1) Add the (@experiment_function) decorator onto your main function (lets call it my_main_fcn).
2) Open up the User Interface by calling my_main_funcion.browse()

There are two ways to interact with the experiment framework:
1) Through the User Interface.
2) Through the experiment API.

In the bottom of this file, you an see example code for running either the UI or API.  You can try the different versions
by either running this file as "python demo_mnist_logreg.py ui" or "python demo_mnist_logreg.py api"
"""


class OnlineLogisticRegressor:

    def __init__(self, n_in, n_out, learning_rate = 0.01):
        self.w = np.zeros((n_in, n_out))
        self.learning_rate = learning_rate

    def train(self, x, targ):  # x: (n_samples, n_in), targ:
        """
        :param x: A (n_samples, n_in) input array
        :param targ: A (n_samples, ) array of interger target labels
        """
        y_pred = self.predict(x)
        dl_dy = y_pred - onehot(targ, n_classes=y_pred.shape[1])
        self.w -= self.learning_rate * (x.T.dot(dl_dy))

    def predict(self, x):  # x: (n_samples, n_in)
        """
        :param x: A (n_samples, n_in) input array
        :return: A (n_samples, n_out) array of output probabilities for each sample.
        """
        return softmax(x.dot(self.w), axis=1)


@experiment_function  # This decorator turns the function into an Experiment object.
def demo_mnist_logreg(minibatch_size=20, learning_rate=0.01, max_training_samples = None, n_epochs=10, test_epoch_period=0.2):
    """
    Train a Logistic Regressor on the MNIST dataset, report training/test scores throughout training, and return the
    final scores.
    """

    x_train, y_train, x_test, y_test = get_mnist_dataset(flat=True, n_training_samples=max_training_samples).xyxy

    predictor = OnlineLogisticRegressor(n_in=784, n_out=10, learning_rate=learning_rate)

    # Train and periodically record scores.
    epoch_scores = []
    for ix, iteration_info in minibatch_index_info_generator(n_samples = len(x_train), minibatch_size=minibatch_size, n_epochs=n_epochs, test_epochs=('every', test_epoch_period)):
        if iteration_info.test_now:
            training_error = 100*(np.argmax(predictor.predict(x_train), axis=1)==y_train).mean()
            test_error = 100*(np.argmax(predictor.predict(x_test), axis=1)==y_test).mean()
            print 'Epoch {epoch}: Test Error: {test}%, Training Error: {train}%'.format(epoch=iteration_info.epoch, test=test_error, train=training_error)
            epoch_scores.append((iteration_info.epoch, training_error, test_error))
        predictor.train(x_train[ix], y_train[ix])

    # Plot
    plt.figure()
    epochs, training_costs, test_costs = zip(*epoch_scores)
    plt.plot(epochs, np.array([training_costs, test_costs]).T)
    plt.xlabel('Epoch')
    plt.ylabel('% Error')
    plt.legend(['Training Error', 'Test Error'])
    plt.title("Learning Curve")
    plt.ylim(80, 100)
    plt.ion()  # Don't hang on plot
    plt.show()

    return {'train': epoch_scores[-1][1], 'test': epoch_scores[-1][2]}  # Return final scores


demo_mnist_logreg.add_variant(learning_rate=0.1)                        # You can define a variant by providing keyword arguments to the function
X=demo_mnist_logreg.add_variant(learning_rate=0.001)                    # You can assign a variant to a variable ...
X.add_variant(minibatch_size=10)                                        # ... which lets you make variants of variants
demo_mnist_logreg.add_variant('small-set', max_training_samples=1000)   # You can optionally give variants names to refer to via experiment.get_variant(name)


if __name__ == '__main__':
    import sys
    demo_version = sys.argv[1] if len(sys.argv) > 1 else 'ui'

    if demo_version == 'ui':
        # Open the experiment browser UI, from where you can run and view experiments:
        demo_mnist_logreg.browse()
        # Commands you can try (or press h to see a list of all commands):
        # run all          # Runs all experiments
        # show 2-3         # Show the output and figures from experiments 2 and 3
        # delete all       # Delete all records
        # delete old       # Delete all records except the newest from each experiment
        # compare 1-3      # Generate a table comparing the arguments and results of experiments 1,2,3

    elif demo_version == 'api':
        # Demonstrate some commands in the api.  Here we will collect the results from several experiments and compare them.

        # Collect a list of demo_mnist_logreg and all its variants.
        exp_list = demo_mnist_logreg.get_all_variants(include_self=True)

        # Run all experiments that have not yet been run to completion:
        for ex in exp_list:
            if len(ex.get_records(only_completed=True))==0:
                ex.run()

        # Print a table containing the full list of arguments and results, and show all figures.
        for ex in exp_list:
            record = ex.get_latest_record(only_completed=True)   # Get an object representing the latest completed run of the experiment
            args = record.get_args()  # This is a list of tuples of (arg_name, arg_value)
            result = record.get_result()  # This is the return value {'train': train score, 'test': test_score}
            (fig, ) = record.load_figures()
            fig.gca().set_title(ex.name)
            print ex.name.ljust(60) \
                  + ' '.join('{}:{}'.format(arg_name, arg_value).ljust(26) for arg_name, arg_value in args) \
                  + '  ||  ' + ' '.join('{}:{}'.format(subset, score).rjust(15) for subset, score in result.items())
        plt.show()
    else:
        raise NotImplementedError('No Demo Version {}'.format(demo_version))



import pytest
from artemis.general.test_mode import set_test_mode
from artemis.plotting.fast import find_interval_extremes, fastplot, fastloglog
import matplotlib.pyplot as plt
import numpy as np
__author__ = 'peter'


@pytest.mark.skipif(True, reason='scipy.weave, which is required, does not reliablly install.')
def test_fastplot():

    plt.ion()
    fastplot(np.random.randn(25000))
    plt.show()
    fastloglog(np.random.rand(25000))
    plt.show()
    fastplot(np.random.randn(1000))
    plt.show()


@pytest.mark.skipif(True, reason='scipy.weave, which is required, does not reliablly install.')
def test_find_interval_extremes():

    arr = np.random.RandomState(324).randn(1000)
    edges = np.linspace(0, 1000, 10)[1:]
    extreme_indices = find_interval_extremes(arr, edges)
    assert len(extreme_indices) <= len(edges)*2
    assert np.std(arr[extreme_indices]) > 2* np.std(arr)  # Good enough test


if __name__ == '__main__':
    set_test_mode(True)
    test_fastplot()
    test_find_interval_extremes()



import re
from collections import OrderedDict

from artemis.experiments.experiment_management import load_lastest_experiment_results
from artemis.experiments.experiment_record import NoSavedResultError, ExpInfoFields, ExperimentRecord
from artemis.experiments.experiments import is_experiment_loadable, GLOBAL_EXPERIMENT_LIBRARY
from artemis.general.display import deepstr, truncate_string, hold_numpy_printoptions, side_by_side
from artemis.general.should_be_builtins import separate_common_items, all_equal, bad_value, izip_equal
from artemis.general.tables import build_table
from tabulate import tabulate


def get_record_result_string(record, func='deep', truncate_to = None, array_print_threshold=8, array_float_format='.3g', oneline=False):
    """
    Get a string representing the result of the experiment.
    :param record:
    :param func:
    :return:
    """
    with hold_numpy_printoptions(threshold = array_print_threshold, formatter={'float': lambda x: '{{:{}}}'.format(array_float_format).format(x)}):
        if isinstance(func, basestring):
            func = {
                'deep': deepstr,
                'str': str,
                }[func]
        else:
            assert callable(func), 'func must be callable'
        try:
            result = record.get_result()
        except NoSavedResultError:
            return '<No result has been saved>'
        string = func(result)

    if truncate_to is not None:
        string = truncate_string(string, truncation=truncate_to, message = '...<truncated>')
    if oneline:
        string = string.replace('\n', ', ')
    return string


def get_record_full_string(record, show_info = True, show_logs = True, truncate_logs = None, show_result ='deep', truncate_result = None):
    """
    Get a human-readable string containing info about the experiment record.

    :param show_info: Show info about experiment (name, id, runtime, etc)
    :param show_logs: Show logs (True, False, or an integer character count to truncate logs at)
    :param show_result: Show the result.  Options for display are:
        'short': Print a one-liner (note: sometimes prints multiple lines)
        'long': Directly call str
        'deep': Use the deepstr function for a compact nested printout.
    :return: A string to print.
    """

    assert show_result in (False, 'full', 'deep')
    full_info_string = '{header} {rid} {header}\n'.format(header="=" * 10, rid=record.get_id())
    if show_info:
        full_info_string += '{}\n'.format(record.info.get_text())
    if show_logs:
        log = record.get_log()
        if truncate_logs is not None and len(log)>truncate_logs:
            log = log[:truncate_logs-100] + '\n\n ... LOG TRUNCATED TO {} CHARACTERS ... \n\n'.format(truncate_logs) + log[-100:]
        full_info_string += '{subborder} Logs {subborder}\n{log}\n'.format(subborder='-' * 20, log=log)
    if show_result:
        result_str = get_record_result_string(record, truncate_to=truncate_result, func=show_result)
        full_info_string += '{subborder} Result {subborder}\n{result}\n'.format(subborder='-' * 20, result=result_str)
    full_info_string += "=" * 50 + '\n'
    return full_info_string


def get_record_invalid_arg_string(record):
    """
    Return a string identifying ig the arguments for this experiment are still valid.
    :return:
    """
    experiment_id = record.get_experiment_id()
    if is_experiment_loadable(experiment_id):
        last_run_args = dict(record.info.get_field(ExpInfoFields.ARGS))
        current_args = dict(record.get_experiment().get_args())
        validity = record.is_valid(last_run_args=last_run_args, current_args=current_args)
        if validity is False:
            last_arg_str, this_arg_str = [
                ['{}:{}'.format(k, v) for k, v in argdict.iteritems()] if isinstance(argdict, dict) else
                ['{}:{}'.format(k, v) for k, v in argdict]
                for argdict in (last_run_args, current_args)]
            common, (old_args, new_args) = separate_common_items([last_arg_str, this_arg_str])
            notes = "No: Args changed!: {{{}}}->{{{}}}".format(','.join(old_args), ','.join(new_args))
        elif validity is None:
            notes = "Cannot Determine."
        else:
            notes = "Yes"
    else:
        notes = "<Experiment Not Currently Imported>"
    return notes


def get_oneline_result_string(record, truncate_to=None, array_float_format='.3g', array_print_threshold=8):
    """
    Get a string that describes the result of the record in one line.  This can optionally be specified by
    experiment.one_liner_function.

    :param record: An ExperimentRecord.
    :param truncate_to:
    :param array_float_format:
    :param array_print_threshold:
    :return: A string with no newlines briefly describing the result of the record.
    """
    if not is_experiment_loadable(record.get_experiment_id()):
        one_liner_function=str
    else:
        one_liner_function = record.get_experiment().one_liner_function
        if one_liner_function is None:
            one_liner_function = str
    return get_record_result_string(record, func=one_liner_function, truncate_to=truncate_to, array_print_threshold=array_print_threshold,
        array_float_format=array_float_format, oneline=True)


def display_experiment_record(record):
    result = record.get_result()
    display_func = record.get_experiment().display_function
    if display_func is None:
        print deepstr(result)
    else:
        display_func(result)


def compare_experiment_results(experiments, error_if_no_result = True):
    comp_functions = [ex.comparison_function for ex in experiments]
    assert all_equal(comp_functions), 'Experiments must have same comparison functions.'
    comp_function = comp_functions[0]
    assert comp_function is not None, 'Cannot compare results, because you have not specified any comparison function for this experiment.  Use @ExperimentFunction(comparison_function = my_func)'
    results = load_lastest_experiment_results(experiments, error_if_no_result=error_if_no_result)
    comp_function(results)


def print_experiment_record_argtable(records):

    # info_results = OrderedDict([(identifier, record.get_info()) for identifier, record in zip(record_identifiers, records)]])

    funtion_names = [record.info.get_field(ExpInfoFields.FUNCTION) for record in records]
    args = [record.info.get_field(ExpInfoFields.ARGS) for record in records]
    results = [record.get_result() for record in records]

    common_args, different_args = separate_common_items(args)

    record_ids = [record.get_id() for record in records]

    def lookup_fcn(record_id, column):
        index = record_ids.index(record_id)
        if column=='Function':
            return funtion_names[index]
        elif column=='Run Time':
            return records[index].info.get_field_text(ExpInfoFields.RUNTIME)
        elif column=='Common Args':
            return ', '.join('{}={}'.format(k, v) for k, v in common_args)
        elif column=='Different Args':
            return ', '.join('{}={}'.format(k, v) for k, v in different_args[index])
        elif column=='Result':
            return results[index]
        else:
            bad_value(column)

    rows = build_table(lookup_fcn,
        row_categories=record_ids,
        column_categories=['Function', 'Run Time', 'Common Args', 'Different Args', 'Result'],
        prettify_labels=False
        )

    print tabulate(rows)


def show_experiment_records(records, parallel_text=None, hang_notice = None, show_logs=True, truncate_logs=10000, truncate_result=10000, show_result ='deep', hang=True):
    """
    Show the console logs, figures, and results of a collection of experiments.

    :param records:
    :param parallel_text:
    :param hang_notice:
    :return:
    """
    if isinstance(records, ExperimentRecord):
        records = [records]
    if parallel_text is None:
        parallel_text = len(records)>1
    if len(records)==0:
        print '... No records to show ...'
    else:
        if parallel_text:
            print side_by_side([get_record_full_string(rec, show_logs=show_logs, show_result=show_result) for rec in records], max_linewidth=128)
        else:  #
            for rec in records:
                print get_record_full_string(rec, show_logs=True, truncate_logs=truncate_logs, show_result=show_result, truncate_result=truncate_result)
    has_matplotlib_figures = any(loc.endswith('.pkl') for rec in records for loc in rec.get_figure_locs())
    from matplotlib import pyplot as plt
    if has_matplotlib_figures:
        from artemis.plotting.saving_plots import interactive_matplotlib_context
        for rec in records:
            rec.show_figures(hang=False)
        if hang_notice is not None:
            print hang_notice

        with interactive_matplotlib_context(not hang):
            plt.show()

    if any(rec.get_experiment().display_function is not None for rec in records):
        from artemis.plotting.saving_plots import interactive_matplotlib_context
        with interactive_matplotlib_context():
            for rec in records:
                display_experiment_record(rec)

    return has_matplotlib_figures


def find_experiment(*search_terms):
    """
    Find an experiment.  Invoke
    :param search_term: A term that will be used to search for an experiment.
    :return:
    """

    found_experiments = OrderedDict((name, ex) for name, ex in GLOBAL_EXPERIMENT_LIBRARY.iteritems() if all(re.search(term, name) for term in search_terms))
    if len(found_experiments)==0:
        raise Exception("None of the {} experiments matched the search: '{}'".format(len(GLOBAL_EXPERIMENT_LIBRARY), search_terms))
    elif len(found_experiments)>1:
        raise Exception("More than one experiment matched the search '{}', you need to be more specific.  Found: {}".format(search_terms, found_experiments.keys()))
    else:
        return found_experiments.values()[0]


def make_record_comparison_table(records, args_to_show=None, results_extractor = None, print_table = False):
    """
    Make a table comparing the arguments and results of different experiment records.  You can use the output
    of this function with the tabulate package to make a nice readable table.

    :param records: A list of records whose results to compare
    :param args_to_show: A list of arguments to show.  If none, it will just show all arguments
        that differ between experiments.
    :param results_extractor: A dict<str->callable> where the callables take the result of the
        experiment as an argument and return an entry in the table.
    :param print_table: Optionally, import tabulate and print the table here and now.
    :return: headers, rows
        headers is a list of of headers for the top of the table
        rows is a list of lists filling in the information.

    example usage:

        headers, rows = make_record_comparison_table(
            record_ids = [experiment_id_to_latest_record_id(eid) for eid in [
                'demo_fast_weight_mlp.multilayer_baseline.1epoch.version=mlp',
                'demo_fast_weight_mlp.multilayer_baseline.1epoch.full-gd.n_steps=1',
                'demo_fast_weight_mlp.multilayer_baseline.1epoch.full-gd.n_steps=20',
                ]],
            results_extractor={
                'Test': lambda result: result.get_best('test').score.get_score('test'),
                'Train': lambda result: result.get_best('test').score.get_score('train'),
                }
             )
        import tabulate
        print tabulate.tabulate(rows, headers=headers, tablefmt=tablefmt)
    """

    args = [rec.info.get_field(ExpInfoFields.ARGS) for rec in records]
    if args_to_show is None:
        common, separate = separate_common_items(args)
        args_to_show = [k for k, v in separate[0]]

    if results_extractor is None:
        results_extractor = {'Result': str}
    elif callable(results_extractor):
        results_extractor = {'Result': results_extractor}
    else:
        assert isinstance(results_extractor, dict)

    headers = args_to_show + results_extractor.keys()

    rows = []
    for record, record_args in izip_equal(records, args):
        arg_dict = dict(record_args)
        args_vals = [arg_dict[k] for k in args_to_show]
        results = record.get_result()
        rows.append(args_vals+[f(results) for f in results_extractor.values()])

    if print_table:
        import tabulate
        print tabulate.tabulate(rows, headers=headers, tablefmt='simple')
    return headers, rows


from artemis.fileman.local_dir import get_local_path
import os

from artemis.fileman.persistent_ordered_dict import PersistentOrderedDict


def test_persistent_ordered_dict():

    file_path = get_local_path('tests/podtest.pkl')
    if os.path.exists(file_path):
        os.remove(file_path)

    with PersistentOrderedDict(file_path) as pod:
        assert pod.items() == []
        pod['a'] = [1, 2, 3]
        pod['b'] = [4, 5, 6]
        pod['c'] = [7, 8]
    pod['d'] = [9, 10]  # Should not be recorded

    with PersistentOrderedDict(file_path) as pod:
        assert pod.items() == [('a', [1, 2, 3]), ('b', [4, 5, 6]), ('c', [7, 8])]
        pod['e']=11

    with PersistentOrderedDict(file_path) as pod:
        assert pod.items() == [('a', [1, 2, 3]), ('b', [4, 5, 6]), ('c', [7, 8]), ('e', 11)]


if __name__ == '__main__':
    test_persistent_ordered_dict()



from decorators import experiment_function, experiment_root, ExperimentFunction
from experiments import capture_created_experiments, load_experiment
from experiment_record import load_experiment_record



import os
import pickle
import re
from contextlib import contextmanager
from datetime import datetime

import numpy as np
from artemis.fileman.file_getter import get_temp_file, get_file_and_cache
from artemis.fileman.images2gif import readGif
from artemis.fileman.local_dir import get_local_path, make_file_dir
from artemis.general.image_ops import get_dark_edge_slice, resize_image


def smart_save(obj, relative_path, remove_file_after = False):
    """
    Save an object locally.  How you save it depends on its extension.
    Extensions currently supported:
        pkl: Pickle file.
        That is all.
    :param obj: Object to save
    :param relative_path: Path to save it, relative to "Data" directory.  The following placeholders can be used:
        %T - ISO time
        %R - Current Experiment Record Identifier (includes experiment time and experiment name)
    :param remove_file_after: If you're just running a test, it's good to verify that you can save, but you don't
        actually want to leave a file behind.  If that's the case, set this argument to True.
    """
    if '%T' in relative_path:
        iso_time = datetime.now().isoformat().replace(':', '.').replace('-', '.')
        relative_path = relative_path.replace('%T', iso_time)
    if '%R' in relative_path:
        from artemis.experiments.experiment_record import get_current_experiment_id
        relative_path = relative_path.replace('%R', get_current_experiment_id())
    _, ext = os.path.splitext(relative_path)

    with smart_file(relative_path, make_dir=True) as local_path:
        print 'Saved object <%s at %s> to file: "%s"' % (obj.__class__.__name__, hex(id(object)), local_path)
        if ext=='.pkl':
            with open(local_path, 'w') as f:
                pickle.dump(obj, f, protocol=pickle.HIGHEST_PROTOCOL)
        elif ext in _IMAGE_EXTENSIONS:
            _save_image(obj, local_path)
        elif ext=='.pdf':
            obj.savefig(local_path)
        else:
            raise Exception("No method exists yet to save '%s' files.  Add it!" % (ext, ))

    if remove_file_after:
        os.remove(local_path)

    return local_path


def smart_load(location, use_cache = False):
    """
    Load a file, with the method based on the extension.  See smart_save doc for the list of extensions.
    :param location: Identifies file location.
        If it's formatted as a url, it's downloaded.
        If it begins with a "/", it's assumed to be a local path.
        Otherwise, it is assumed to be referenced relative to the data directory.
    :param use_cache: If True, and the location is a url, make a local cache of the file for future use (note: if the
        file at this url changes, the cached file will not).
    :return: An object, whose type depends on the extension.  Generally a numpy array for data or an object for pickles.
    """
    assert isinstance(location, str), 'Location must be a string!  We got: %s' % (location, )
    with smart_file(location, use_cache=use_cache) as local_path:
        ext = os.path.splitext(local_path)[1].lower()
        if ext=='.pkl':
            with open(local_path) as f:
                obj = pickle.load(f)
        elif ext=='.gif':
            frames = readGif(local_path)
            if frames[0].shape[2]==3 and all(f.shape[2] for f in frames[1:]):  # Wierd case:
                obj = np.array([frames[0]]+[f[:, :, :3] for f in frames[1:]])
            else:
                obj = np.array(readGif(local_path))
        elif ext in _IMAGE_EXTENSIONS:
            from PIL import Image
            obj = _load_image(local_path)
        elif ext in ('.mpg', '.mp4', '.mpeg'):
            obj = _load_video(local_path)
        else:
            raise Exception("No method exists yet to load '%s' files.  Add it!" % (ext, ))
    return obj


def smart_load_image(location, max_resolution = None, force_rgb=False, use_cache = False):
    """
    Load an image into a numpy array.

    :param location: Identifies file location.
        If it's formatted as a url, it's downloaded.
        If it begins with a "/", it's assumed to be a local path.
        Otherwise, it is assumed to be referenced relative to the data directory.
    :param max_resolution: Maximum resolution (size_y, size_x) of the image
    :param force_rgb: Force an RGB representation (transform greyscale and RGBA images into RGB)
    :param use_cache: If True, and the location is a url, make a local cache of the file for future use (note: if the
        file at this url changes, the cached file will not).
    :return: An object, whose type depends on the extension.  Generally a numpy array for data or an object for pickles.
    """
    with smart_file(location, use_cache=use_cache) as local_path:
        return _load_image(local_path, max_resolution = max_resolution, force_rgb=force_rgb)

_IMAGE_EXTENSIONS = ('.jpg', '.jpeg', '.png', '.gif')


def _load_image(local_path, max_resolution = None, force_rgb = False):
    """
    :param local_path: Local path to the file
    :param max_resolution: Maximum resolution (size_y, size_x) of the image
    :param force_rgb: Force an RGB representation (transform greyscale and RGBA images into RGB)
    :return: The image array, which can be:
        (size_y, size_x, 3) For a colour RGB image
        (size_y, size_x) For a greyscale image
        (size_y, size_x, 4) For an image with transperancy (Note: Disabled for now)
    """
    # TODO: Consider replacing PIL with scipy
    ext = os.path.splitext(local_path)[1].lower()
    assert ext in _IMAGE_EXTENSIONS, "Can't deal with extension: {}".format(ext)
    from PIL import Image
    pic = Image.open(local_path)

    if max_resolution is not None:
        max_width, max_height = max_resolution
        pic.thumbnail((max_width, max_height), Image.ANTIALIAS)

    pic_arr = np.asarray(pic)

    if force_rgb:
        if pic_arr.ndim==2:
            pic_arr = np.repeat(pic_arr[:, :, None], 3, axis=2)
        elif pic_arr.shape[2]==4:
            pic_arr = pic_arr[:, :, :3]
        else:
            assert pic_arr.shape[2]==3
    return pic_arr


def _save_image(image_array, local_path):
    ext = os.path.splitext(local_path)[1].lower()
    assert ext in _IMAGE_EXTENSIONS
    from PIL import Image
    pic = Image.fromarray(image_array)
    pic.save(local_path)


@contextmanager
def smart_file(location, use_cache = False, make_dir = False):
    """
    :param location: Specifies where the file is.
        If it's formatted as a url, it's downloaded.
        If it begins with a "/", it's assumed to be a local path.
        Otherwise, it is assumed to be referenced relative to the data directory.
    :param use_cache: If True, and the location is a url, make a local cache of the file for future use (note: if the
        file at this url changes, the cached file will not).
    :param make_dir: Make the directory for this file, if it does not exist.
    :yield: The local path to the file.
    """
    its_a_url = is_url(location)
    if its_a_url:
        assert not make_dir, "We cannot 'make the directory' for a URL"
        if use_cache:
            local_path = get_file_and_cache(location)
        else:
            local_path = get_temp_file(location)
    else:
        local_path = get_local_path(location)
        if make_dir:
            make_file_dir(local_path)

    yield local_path

    if its_a_url and not use_cache:
        os.remove(local_path)


def is_url(path):
    regex = re.compile(
        r'^(?:http|ftp)s?://' # http:// or https://
        r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|' #domain...
        r'localhost|' #localhost...
        r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})' # ...or ip
        r'(?::\d+)?' # optional port
        r'(?:/?|[/?]\S+)$', re.IGNORECASE)
    return True if re.match(regex, path) else False


def smart_load_video(location, use_cache = False, resize_mode='resize_and_crop', cut_edges=False, size = None, cut_edges_thresh=0):
    """
    :param location:
    :param size: A 2-tuple of width-height, indicating the desired size of the ouput
    :param resize_mode: The mode with which to get the video to the desired size.  Can be:
        'squeeze', 'preserve_aspect', 'crop', 'scale_crop'.  See resize_image in image_ops.py for more info.
    :param cut_edges: True if you want to cut the dark edges from the video
    :param cut_edges_thresh: If cut_edges, this is the threshold at which you'd like to cut them.
    :return: A (n_frames, height, width, 3) numpy array
    """

    with smart_file(location, use_cache=use_cache) as local_path:
        return _load_video(local_path, resize_mode=resize_mode, cut_edges=cut_edges, size=size, cut_edges_thresh=cut_edges_thresh)


def _load_video(full_path, size = None, resize_mode = 'resize_and_crop', cut_edges=False, cut_edges_thresh=0):
    """
    Lead a video into a numpy array

    :param full_path: Full path to the video
    :param size: A 2-tuple of width-height, indicating the desired size of the ouput
    :param resize_mode: The mode with which to get the video to the desired size.  Can be:
        'squeeze', 'preserve_aspect', 'crop', 'scale_crop'.  See resize_image in image_ops.py for more info.
    :param cut_edges: True if you want to cut the dark edges from the video
    :param cut_edges_thresh: If cut_edges, this is the threshold at which you'd like to cut them.
    :return: A (n_frames, height, width, 3) numpy array
    """
    try:
        from moviepy.video.io.VideoFileClip import VideoFileClip
    except ImportError:
        raise ImportError("You need to install moviepy to read videos.  In the virtualenv, go `pip install moviepy`")
    assert os.path.exists(full_path)
    video = VideoFileClip(full_path)
    images = []
    edge_crops = None
    for frame in video.iter_frames():
        if cut_edges:
            if edge_crops is None:
                edge_crops = get_dark_edge_slice(frame, cut_edges_thresh=cut_edges_thresh)
            else:
                frame = frame[edge_crops[0], edge_crops[1]]
        if size is not None:
            width, height = size
            frame = resize_image(frame, width=width, height=height, mode=resize_mode)
        images.append(frame)
    return images



import matplotlib
import pytest
matplotlib.use('TkAgg')
import matplotlib.pyplot as plt
import numpy as np
__author__ = 'peter'

"""
Small demo that shows how to get pixel values of a plot.
"""
# TODO: Figure out how to save these as a GIF or some animation.


@pytest.mark.skipif(True, reason = "This fails in pytest but runs when run directly.  I don't know why, but it doesn't matter for now.")
def test_animation():

    fig = plt.figure()
    get_data = lambda: np.random.randn(5, 6)
    plt.ion()
    h=plt.imshow(get_data())
    plt.show()

    frames = []
    for i in xrange(20):
        lastframe = fig2im(fig)
        frames.append(fig2im(fig))
        h.set_array(lastframe)
        plt.draw()


def fig2im(fig):
    """
    @brief Convert a Matplotlib figure to a 3D numpy array with RGB channels and return it
    @param fig a matplotlib figure
    @return a numpy 3D array of RGB values
    """
    fig.canvas.draw()

    # Get the RGBA buffer from the figure
    w, h = fig.canvas.get_width_height()
    buf = np.fromstring ( fig.canvas.tostring_rgb(), dtype=np.uint8)
    buf.shape = (h, w, 3)

    return buf


if __name__ == '__main__':
    test_animation()




import numpy as np
import pytest
from artemis.ml.predictors.learning_curve_plots import plot_learning_curves
from artemis.ml.datasets.synthetic_clusters import get_synthetic_clusters_dataset
from artemis.ml.predictors.i_predictor import IPredictor
from artemis.general.mymath import sqrtspace
from artemis.ml.predictors.perceptron import Perceptron
from artemis.ml.predictors.predictor_comparison import compare_predictors, assess_online_predictor


__author__ = 'peter'


@pytest.mark.skip(True, reason = 'Test uses sklearn, which should not be a requirement for artemis.  It does work though.')
def test_compare_predictors(hang_plot = False):
    from sklearn.svm import SVC
    dataset = get_synthetic_clusters_dataset()

    w_constructor = lambda rng = np.random.RandomState(45): .1*rng.randn(dataset.input_shape[0], dataset.n_categories)
    records = compare_predictors(
        dataset = dataset,
        offline_predictors={
            'SVM': SVC()
            },
        online_predictors={
            'fast-perceptron': Perceptron(alpha = 0.1, w = w_constructor()).to_categorical(),
            'slow-perceptron': Perceptron(alpha = 0.001, w = w_constructor()).to_categorical()
            },
        minibatch_size = 10,
        test_epochs = sqrtspace(0, 10, 20),
        evaluation_function='percent_correct'
        )

    assert 99 < records['SVM'].get_scores('Test') <= 100
    assert 20 < records['slow-perceptron'].get_scores('Test')[0] < 40 and 95 < records['slow-perceptron'].get_scores('Test')[-1] <= 100
    assert 20 < records['fast-perceptron'].get_scores('Test')[0] < 40 and 98 < records['fast-perceptron'].get_scores('Test')[-1] <= 100

    plot_learning_curves(records, hang = hang_plot)


def test_stretch_minibatches():
    """
    Assert that the 'stretch' minibatch size works as expected.
    """
    class _SingleSampleRegressor(IPredictor):

        def __init__(self, n_in, n_out, eta = 0.1):
            self.w = np.zeros((n_in, n_out))
            self.eta = eta
            self.n_train_calls = 0

        def _predict_one(self, x):
            return x.dot(self.w)

        def _train_one(self, x, y):
            self.w += self.eta * np.outer(x, y)

        def predict(self, x):
            return np.array([self._predict_one(xi) for xi in x])

        def train(self, x, y):
            for xi, yi in zip(x, y):
                self._train_one(xi, yi)
            self.n_train_calls += 1

    dataset = get_synthetic_clusters_dataset().to_onehot()

    p1 = _SingleSampleRegressor(dataset.input_size, dataset.target_size)
    p1_scores = assess_online_predictor(p1, dataset, evaluation_function='percent_argmax_correct', test_epochs=[0, .5, 1], minibatch_size=1, test_on = 'test')
    p1_trained_out = p1.predict(dataset.test_set.input)
    assert p1.n_train_calls == dataset.training_set.n_samples
    scores = p1_scores.get_scores()
    assert len(scores) == 3 and scores[0] < 28 and scores[2] > 99

    p2 = _SingleSampleRegressor(dataset.input_size, dataset.target_size)
    assess_online_predictor(p2, dataset, evaluation_function='percent_argmax_correct', test_epochs=[0, 0.5, 1], minibatch_size='stretch',  test_on = 'test')
    p2_trained_out = p2.predict(dataset.test_set.input)
    assert p2.n_train_calls == 2

    assert np.array_equal(p1_trained_out, p2_trained_out)


if __name__ == '__main__':
    test_compare_predictors(hang_plot=True)
    test_stretch_minibatches()



import logging
import os
from functools import partial
from artemis.fileman.local_dir import get_local_path, make_file_dir
from artemis.general.functional import infer_arg_values
from artemis.general.hashing import compute_fixed_hash
from artemis.general.test_mode import is_test_mode

logging.basicConfig()
LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.INFO)


__author__ = 'peter'

MEMO_WRITE_ENABLED = True
MEMO_READ_ENABLED = True
MEMO_DIR = get_local_path('memoize_to_disk')


def memoize_to_disk(fcn, local_cache = False, disable_on_tests=True, use_cpickle = False):
    """
    Save (memoize) computed results to disk, so that the same function, called with the
    same arguments, does not need to be recomputed.  This is useful if you have a long-running
    function that is often being given the same arguments.  Note: this does NOT check for the state
    of Global variables/time/whatever else the function may use, so you need to make sure your
    function is truly a function in that outputs only depend on inputs.  Otherwise, this will
    give you misleading results.

    Usage:
        @memoize_to_disk
        def fcn(a, b, c = None):
            results = ...
            return results

    You can also use this without the decorator.
    e.g.
        result = memoize_to_disk(fcn)(a, b, c=3)

    This is useful if:
    a) The decorator can/should not be visible from where the function is defined.
    b) You only want to memoize the function in one use-case, but not all.

    :param fcn: The function you're decorating
    :param local_cache: Keep a cache in python (so you don't need to go to disk if you call again in the same process)
    :param disable_on_tests: Persistent memos can really screw up tests, so disable memos when is_test_mode() returns
        True.  Generally, leave this as true, unless you are testing memoization itself.
    :param use_cpickle: Use CPickle, instead of pickle, to save results.  This can be faster for complex python
        structures, but can be slower for numpy arrays.  So we recommend not using it.
    :return: A wrapper around the function that checks for memos and loads old results if they exist.
    """

    if use_cpickle:
        import cPickle as pickle
    else:
        import pickle

    cached_local_results = {}

    def check_memos(*args, **kwargs):

        if disable_on_tests and is_test_mode():
            return fcn(*args, **kwargs)

        result_computed = False
        full_args = infer_arg_values(fcn, *args, **kwargs)
        filepath = get_function_hash_filename(fcn, full_args)
        # The filepath is used as the unique identifier, for both the local path and the disk-path
        # It may be more efficient to use the built-in hashability of certain types for the local cash, and just have special
        # ways of dealing with non-hashables like lists and numpy arrays - it's a bit dangerous because we need to check
        # that no object or subobjects have been changed.
        if MEMO_READ_ENABLED:
            if local_cache:
                # local_cache_signature = get_local_cache_signature(args, kwargs)
                if filepath in cached_local_results:
                    LOGGER.info('Reading disk-memo from local cache for function %s' % (fcn.__name__, ))
                    return cached_local_results[filepath]
            if os.path.exists(filepath):
                with open(filepath) as f:
                    try:
                        LOGGER.info('Reading memo for function %s' % (fcn.__name__, ))
                        result = pickle.load(f)
                    except (ValueError, ImportError) as err:
                        if isinstance(err, ValueError):
                            LOGGER.warn('Memo-file "%s" was corrupt.  (%s: %s).  Recomputing.' % (filepath, err.__class__.__name__, str(err)))
                        elif isinstance(err, ImportError):
                            LOGGER.warn('Memo-file "{}" was tried to reference an old class and got ImportError: {}.  Recomputing.'.format(filepath, str(err)))
                        result_computed = True
                        result = fcn(*args, **kwargs)
            else:
                result_computed = True
                result = fcn(*args, **kwargs)
        else:
            result_computed = True
            result = fcn(*args, **kwargs)

        if MEMO_WRITE_ENABLED and result is not None:  # We assume result of None means you haven't done coding your function.
            if local_cache:
                cached_local_results[filepath] = result
            if result_computed:  # Result was computed, so write it down
                filepath = get_function_hash_filename(fcn, full_args)
                make_file_dir(filepath)
                with open(filepath, 'w') as f:
                    LOGGER.info('Writing disk-memo for function %s' % (fcn.__name__, ))
                    pickle.dump(result, f, protocol=2)

        return result

    check_memos.wrapped_fcn = fcn
    check_memos.clear_cache = lambda: clear_memo_files_for_function(check_memos)

    return check_memos


def memoize_to_disk_test(fcn):
    """
    Use this just when testing the memoization itself (because normally memoization is disabled when is_test_mode() is True.
    :param fcn:
    :return:
    """
    return memoize_to_disk(fcn, disable_on_tests=False)


def memoize_to_disk_and_cache(fcn):
    """
    Memoize to disk AND keep a local cache (as you would with the @memoize decorator).
    """
    return memoize_to_disk(fcn, local_cache=True)


def memoize_to_disk_and_cache_test(fcn):
    return memoize_to_disk(fcn, local_cache=True, disable_on_tests=False)


def get_function_hash_filename(fcn, argname_argvalue_list):
    args_code = compute_fixed_hash(argname_argvalue_list)
    # TODO: Include function path in hash?  Or module path, which would allow memos to be shareable.
    return os.path.join(MEMO_DIR, '%s-%s.pkl' % (fcn.__name__, args_code))


def memoize_to_disk_with_settings(**kwargs):
    return partial(memoize_to_disk, **kwargs)


def get_all_memos():
    """
    :return: A list of file-locations
    """
    all_memos = os.listdir(MEMO_DIR) if os.path.exists(MEMO_DIR) else []
    full_paths_of_memos = [os.path.join(MEMO_DIR, m) for m in all_memos]
    return full_paths_of_memos


def get_memo_files_for_function(fcn):
    all_memos = os.listdir(MEMO_DIR) if os.path.exists(MEMO_DIR) else []
    matching_memos = [os.path.join(MEMO_DIR, m) for m in all_memos if m.startswith(fcn.wrapped_fcn.__name__)]
    return matching_memos


def clear_memo_files_for_function(fcn):
    memos = get_memo_files_for_function(fcn)
    for m in memos:
        os.remove(m)


def clear_all_memos():
    all_memos = get_all_memos()
    for m in all_memos:
        os.remove(m)
    print 'Removed %s memos.' % (len(all_memos))


class DisableMemoReading(object):

    def __enter__(self):
        global MEMO_READ_ENABLED
        self._old_read_state = MEMO_READ_ENABLED
        MEMO_READ_ENABLED = False

    def __exit__(self, *args):
        global MEMO_READ_ENABLED
        MEMO_READ_ENABLED = self._old_read_state


class DisableMemoWriting(object):

    def __enter__(self):
        global MEMO_WRITE_ENABLED
        self._old_write_state = MEMO_WRITE_ENABLED
        MEMO_WRITE_ENABLED = False

    def __exit__(self, *args):
        global MEMO_WRITE_ENABLED
        MEMO_WRITE_ENABLED = self._old_write_state


class DisableMemos(object):
    """
    You can disable memoization with a with.

    with DisableMemos():
        # call memoized function, but disable reading/writing.
        my_memoized_function()
    """

    def __enter__(self):
        self._reader = DisableMemoReading()
        self._writer = DisableMemoWriting()
        self._reader.__enter__()
        self._writer.__enter__()

    def __exit__(self, *args):
        self._reader.__exit__(*args)
        self._writer.__exit__(*args)


if __name__ == '__main__':

    cmd = raw_input('Type "clearall" to clear all memos: ')

    if cmd == 'clearall':
        clear_all_memos()
    else:
        raise Exception('Bad command or file name.')




class TraceCollection(object):

    def __init__(self):
        self._traces = {}

    def record(self, value, name):
        if name not in self._traces:
            self._traces[name] = []
        self._traces[name].append(value)

    def get_trace(self, name):
        return self._traces[name]
